import unittest
import numpy as np
from astropy import units as u
from astropy import constants
from pst import dust, SSP

class TestDust(unittest.TestCase):

    @classmethod
    def setUpClass(self):
        print("Setting SSP model for testing dust models")
        self.dummy_wavelength = np.geomspace(100, 1e5, 3000) * u.angstrom
        # Monocromatic SED
        self.dummy_spectra = np.ones(
            self.dummy_wavelength.size) * constants.c / self.dummy_wavelength**2 * 3631 * u.Jy

        self.ssp_model = SSP.PopStar(IMF="cha")

    def test_dust_screen(self):
        dust_screen = dust.DustScreen("ccm89")
        
        ext = dust_screen.get_extinction(self.dummy_wavelength)
        self.assertTrue(np.isfinite(ext).all())
        emission = dust_screen.get_emission(self.dummy_wavelength)
        self.assertTrue(np.allclose(emission, 0))

        ext_spectra = dust_screen.apply_extinction(
            self.dummy_wavelength, self.dummy_spectra)
        self.assertTrue(np.isfinite(ext_spectra).all())
        ext_em_spectra = dust_screen.apply_emission(
            self.dummy_wavelength, ext_spectra)
        self.assertTrue(np.allclose(ext_spectra, ext_em_spectra))

        reddened_ssp = dust_screen.redden_ssp_model(self.ssp_model, a_v=1.0)


if __name__ == '__main__':
    unittest.main()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Jun  9 12:40:50 2025

@author: pcorchoc
"""

from pst.SSP import PopStar
from pst.observables import load_photometric_filters
import numpy as np
from matplotlib import pyplot as plt

from pst.observables import EquivalentWidth

# Define the IMF and redshift
imf_type = 'cha'  # Chabrier 2003 Initial Mass Function
z_redshift = 0.0  # Redshift at which the SSPs are observed

# Load photometric filters (g and r bands)
list_of_filter_names = ["SLOAN_SDSS.g", "SLOAN_SDSS.r"]
filters = load_photometric_filters(list_of_filter_names)

# Initialize the SSP model
ssp_model = PopStar(IMF=imf_type)

# Compute photometry for the SSPs in the selected filters
photometry_fluxes = ssp_model.compute_photometry(filter_list=filters,
                                                 z_obs=z_redshift)

# Convert fluxes to AB magnitudes
magnitudes = -2.5 * np.log10(photometry_fluxes.to_value("3631 Jy / Msun"))

# Compute g - r color
color = magnitudes[0] - magnitudes[1]

# Plot the color as a function of stellar age and metallicity
plt.figure()
plt.pcolormesh(np.log10(ssp_model.ages.value),
               np.log10(ssp_model.metallicities.value), color,
               cmap='jet')
plt.colorbar(label=r'$g - r$')
plt.xlabel('log(Age / yr)')
plt.ylabel('log(Z / Z_sun)')
plt.title(r'$g - r$ Color as a Function of Age and Metallicity')

plt.savefig("source/_static/images/gr_popstar.png", dpi=250, bbox_inches="tight")
plt.close()
# plt.show()

### USER GUIDE

from pst.models import ExponentialDelayedCEM
# Create a model based on a delayed-tau exponential SFH model
# with constant metallicity
cem_model = ExponentialDelayedCEM(mass_today=1.0, today=13.7, tau=3.0,
ism_metallicity_today=0.02)

import numpy as np
from astropy import units as u 
from matplotlib import pyplot as plt

cosmic_time = np.arange(0, 13.7) * u.Gyr
mass_formation_history = cem_model.stellar_mass_formed(cosmic_time)

plt.figure()
plt.plot(cosmic_time, mass_formation_history)
plt.xlabel('Cosmic Time (Gyr)')
plt.ylabel('Stellar Mass Formed (M$_\odot$)')
plt.savefig("source/_static/images/delayed_tau_exponential_sfh.png", dpi=300, bbox_inches='tight')
plt.close()
# plt.show()


sed = cem_model.compute_SED(ssp_model, t_obs=13.7 * u.Gyr)

plt.figure()
plt.loglog(ssp_model.wavelength, sed)
plt.xlabel('Wavelength (Angstrom)')
plt.ylabel('SED (Lsun/Angstrom)')
plt.title('SED of the Delayed Tau Exponential SFH Model')
plt.savefig("source/_static/images/delayed_tau_exponential_sfh_sed.png", dpi=300, bbox_inches='tight')
# plt.show()
plt.close()
# Compute the photometry for the SED in the selected filters


from pst.observables import Filter

# Load a JWST filter from the SVO Filter Service
jwst_miri_filter = Filter.from_svo("JWST_MIRI.F2550W")

# Compute a synthetic magnitude from the galaxy SED computed above
jwst_miri_filter.interpolate(ssp_model.wavelength)
print(sed)
ab_mag = jwst_miri_filter.get_ab(sed / 4 / np.pi / (10 * u.pc)**2)
print(ab_mag)


custom_ew = EquivalentWidth(left_wl_range=[4827.875, 4847.875],
                            central_wl_range=[4847.875, 4876.625],
                            right_wl_range=[4876.625, 4891.625])
np.random.seed(42)  # For reproducibility
noisy_sed = np.random.normal(sed, 0.01 * sed, size=sed.shape) << sed.unit # Simulate some noise in the SED
ew, ew_err = custom_ew.compute_ew(wavelength=ssp_model.wavelength,
                                  spectra=noisy_sed, spectra_err=sed * 0.01)

print(ew, ew_err)

plt.figure()
plt.title(r"EW(H$\beta$)=" + f'{ew.value:.2f} +/- {ew_err:.2f}')
plt.errorbar(ssp_model.wavelength.value, noisy_sed.value, yerr=sed.value * 0.01, label='SED')
plt.axvspan(*custom_ew.central_wl_range.value, color='green', label='Central window', alpha=0.3)
plt.axvspan(*custom_ew.right_wl_range.value, color='red', alpha=0.3, label='Left Range')
plt.axvspan(*custom_ew.left_wl_range.value, color='blue', alpha=0.3, label='Right Range')
plt.xlim(custom_ew.left_wl_range[0].value - 100, custom_ew.right_wl_range[-1].value + 100)
plt.ylim(np.interp(4861 << u.angstrom, ssp_model.wavelength, sed).value * np.array([0.8, 1.2]))
plt.xlabel('Wavelength (Angstrom)')
plt.ylabel('Flux (Lsun/Angstrom)')
plt.legend()
plt.savefig("source/_static/images/ew_calculation.png", dpi=300, bbox_inches='tight')
plt.show()


from abc import ABC, abstractmethod

import numpy as np
from astropy import units as u
from scipy import special
from scipy import interpolate

from pst.SSP import SSPBase
from pst.utils import check_unit, SQRT_2


# CEM utils and mixins

class MassPropMetallicityMixin:
    r"""Model mixin where the metallicity is proportional to the stellar mass
    
    .. math::
        Z(t) = Z_{\rm today} \cdot \left(\frac{M_{\star}(t)}{M_{\star}(\rm today)}\right)^\alpha
        :label: powlaw-metallicity

    where :math:`Z_{\rm today}` is the ISM metallicity at present,
    :math:`M_{\star}(t)` is the stellar mass formed at time :math:`t`, and
    :math:`M_{\star}(\rm today)` is the stellar mass at present.
    The exponent :math:`\alpha` regulates the chemical enrichment rate with stellar mass.

    """
    @property
    def ism_metallicity_today(self):
        """ISM metals mass fraction at present."""
        return self._ism_metallicity_today
    
    @ism_metallicity_today.setter
    def ism_metallicity_today(self, value):
        self._ism_metallicity_today = value
    
    @property
    def alpha_powerlaw(self):
        """Stellar mass power-law exponent."""
        return self._alpha_powerlaw

    @alpha_powerlaw.setter
    def alpha_powerlaw(self, value):
        self._alpha_powerlaw = value

    def ism_metallicity(self, times):
        """Evaluate the ISM metallicity at a given set of times, see eq. :eq:`powlaw-metallicity`.

        Parameters
        ----------
        times : :class:`astropy.units.Quantity`
            Cosmic times at which the metallicity will be evaluated.

        Returns
        -------
        z_t : :class:`astropy.units.Quantity`
            Vector with the ISM metallicity at each input time.
        """
        m = self.stellar_mass_formed(times)
        return self.ism_metallicity_today * np.power(m / self.mass_today, self.alpha_powerlaw)

def sfh_quenching_decorator(stellar_mass_formed):
    """A decorator for including a quenching event in a given SFH."""
    def wrapper_stellar_mass_formed(*args):
        quenching_time = getattr(args[0], "quenching_time", 20.0 << u.Gyr)
        stellar_mass = stellar_mass_formed(*args)
        final_mass = stellar_mass_formed(args[0], quenching_time)
        return np.where(args[1] < quenching_time, stellar_mass, final_mass)
    return wrapper_stellar_mass_formed


class ChemicalEvolutionModel(ABC):
    """
    Abstract base class for chemical evolution models.

    This class provides an interface for modeling the chemical and stellar
    evolution of a galaxy over time. It includes methods for computing the 
    Spectral Energy Distribution (SED), stellar mass, and photometry from 
    a given Simple Stellar Population (SSP) model. The specific methods for 
    computing the star formation rate (SFR) and the metallicity evolution 
    (Z-SFR) need to be implemented in a subclass.
    """
    
    @property
    def today(self):
        return self._today

    @today.setter
    def today(self, value):
        if value is not None:
            self._today = check_unit(value, u.Gyr)
        else:
            self._today = None

    def __init__(self, **kwargs):
        self.today = kwargs.get("today")

    @abstractmethod
    def stellar_mass_formed(self, time) -> u.Quantity:
        """Total stellar mass formed at a given time."""
        return

    @abstractmethod
    def ism_metallicity(self, time) -> u.Quantity:
        """ISM metals mass fraction at a given time."""
        return

    @u.quantity_input
    def interpolate_ssp_masses(self, ssp: SSPBase, t_obs: u.Gyr,
                               oversample_factor=10) -> u.Quantity:
        """
        Interpolate the star formation history to compute the SSP stellar masses.

        This method computes the star formation history of a galaxy over time
        and uses it to interpolate the stellar masses for a given Simple Stellar
        Population (SSP) model at the time of observation.

        Parameters
        ----------
        SSP : :class:`pst.SSP.SSPBase`
            The Simple Stellar Population (SSP) model used for synthesizing the SED.
        t_obs : astropy.Quantity
            The cosmic time at which the galaxy is observed. Only SSPs with ages
            younger than `t_obs` are used.
        oversample_factor : int
            Ages oversampling factor. 

        Returns
        -------
        weights : astropy.Quantity
            Stellar masses corresponding to each SSP age and metallicity, in units
            of solar masses.
        """
        # define age bins from 0 to t_obs
        age_bins = np.hstack(
            [0 << u.yr, np.sqrt(ssp.ages[1:] * ssp.ages[:-1]), 1e12 << u.yr])
        age_bins = age_bins[:age_bins.searchsorted(t_obs) + 1]
        age_bins[-1] = t_obs
        # oversample
        w1 = np.arange(oversample_factor) / oversample_factor
        age_bins = np.hstack(
            [(1-w1) * age_bins[i] + w1 * age_bins[i + 1] for i in range(age_bins.size - 1)]
            + [t_obs])

        # find bin properties
        mass = self.stellar_mass_formed(t_obs - age_bins)
        bin_mass = mass[:-1] - mass[1:]
        bin_age = (age_bins[1:] + age_bins[:-1]) / 2
        bin_metallicity = self.ism_metallicity(t_obs - bin_age)
 
        return ssp.get_weights(ages=bin_age,
                               metallicities=bin_metallicity,
                               masses=bin_mass)

    def time_at_stellar_mass_frac(self, frac, today=None, time_res=30 << u.Myr):
        """
        Get the cosmic time at which the model formed a fraction of the total stellar mass.

        Parameters
        ----------
        frac : float
            Input fraction.
        today : float or astropy.units.Quantity
            Age of the Universe at the time of the observation. If not provided,
            the defult is to use ``self.today'' (returning an error if not set).
        time_res : float or astropy.units.Quantity
            Time-step resolution for sampling the SFH.
        Returns
        -------
        frac_times : astropy.units.Quantity
            Fraction formation times.
        """
        frac = np.atleast_1d(frac)
        if today is None:
            if self.today is None:
                raise ValueError(
                    "current CEM does not have attribute 'today' set,"
                    + " it must be provided by the user")
            else:
                today = self.today
        else:
            today = check_unit(today, u.Gyr)

        time_res = check_unit(time_res, u.Gyr)
        dummy_time = np.arange(0, today.to_value("Gyr"),
                               time_res.to_value("Gyr")) << u.Gyr
        mass_history = self.stellar_mass_formed(dummy_time)
        frac_history = mass_history / mass_history[-1]
        idx = np.searchsorted(frac_history, frac).clip(
            min=1, max=frac_history.size - 1)
        t_frac = (
            dummy_time[idx - 1] * (frac - frac_history[idx]) / (frac_history[idx] - frac_history[idx - 1])
         + dummy_time[idx] * (1 - (frac - frac_history[idx]) / (frac_history[idx] - frac_history[idx - 1])
        ))
        return t_frac

    def compute_SED(self, ssp : SSPBase, t_obs : u.Quantity,
                    allow_negative=False) -> u.Quantity:
        """
        Compute the Spectral Energy Distribution (SED) resulting from the SFH.

        This method synthesizes the SED resulting from the chemical evolution model,
        observed at a given time, using the provided SSP model.

        Parameters
        ----------
        SSP : pst.SSP.SSPBase
            The Simple Stellar Population (SSP) model used for synthesizing the SED.
        t_obs : astropy.Quantity
            The cosmic time at which the galaxy is observed.
        allow_negative : bool, optional
            Whether to allow SSPs with negative masses in the SED computation.
            Default is True.

        Returns
        -------
        sed : astropy.Quantity
            The spectral energy distribution, in the same units as `SSP.L_lambda`.
        
        See also
        --------
        :func:`interpolate_ssp_masses`
        """
        weights = self.interpolate_ssp_masses(ssp, t_obs)
        if not allow_negative:
            mask = (weights > 0) & np.isfinite(weights)
        else:
            mask = np.isfinite(weights)
        sed = np.sum(weights[mask, np.newaxis] * ssp.L_lambda[mask, :],
                    axis=(0))
        return sed

    def compute_photometry(self, ssp, t_obs, photometry=None) -> u.Quantity:
        """
        Compute the synthetic photometry using a SSP at a given time.

        This method computes the photometric fluxes associated to
        the SFH by synthesizing the fluxes of the input SSP model.

        Parameters
        ----------
        ssp : pst.SSP.SSPBase
            The Simple Stellar Population (SSP) model used for generating the
            synthetic photometry.
        t_obs : astropy.Quantity
            The cosmic time at which the galaxy is observed.
        photometry : np.ndarray, optional
            A grid of luminosities in multiple photometric bands. If None, the 
            default SSP photometry will be used. The last two dimensions must 
            match the metallicity and age grid of the SSP model.

        Returns
        -------
        model_photometry : astropy.Quantity
            The photometry of the galaxy in the same units as the input photometry.
        """
        weights = self.interpolate_ssp_masses(ssp, t_obs)
        if photometry is None:
            photometry = ssp.photometry
        if not isinstance(photometry, u.Quantity):
            print("Assuming input photometry array in Jy/Msun")
            photometry *= u.Jy / u.Msun
        extra_dim = photometry.ndim - weights.ndim
        if extra_dim > 0:
            new_dims = tuple(np.arange(extra_dim))
            np.expand_dims(weights, new_dims)
        
        model_photometry = np.sum(photometry * weights, axis=(-1, -2))
        return model_photometry


#-------------------------------------------------------------------------------
class SingleBurstCEM(ChemicalEvolutionModel):
    """
    Single-burst star formation model.

    This class models a galaxy's star formation history as a single burst
    occurring at a specific time, after which no further star formation occurs.

    Attributes
    ----------
    mass_burst : astropy.Quantity
        Total stellar mass formed in the burst.
    time_burst : astropy.Quantity
        Time of the starburst in cosmic time.
    burst_metallicity : astropy.Quantity
        Metallicity of the burst.
    """
    def __init__(self, **kwargs):
        self.mass_burst = check_unit(kwargs['mass_burst'], u.Msun)
        self.time_burst = check_unit(kwargs['time_burst'], u.Gyr)
        self.burst_metallicity = kwargs.get("burst_metallicity",
                                            0.02 * u.dimensionless_unscaled)
        ChemicalEvolutionModel.__init__(self, **kwargs)

    @u.quantity_input
    def stellar_mass_formed(self, time : u.Gyr):
        """Total stellar mass formed at a given time."""
        mass = np.zeros(time.size, dtype=float) * self.mass_burst.unit
        mass[time >= self.time_burst] = self.mass_burst
        return mass

    @u.quantity_input
    def ism_metallicity(self, time : u.Gyr):
        """ISM metals mass fraction at a given time."""
        return np.full(time.size, fill_value=self.burst_metallicity)


#-------------------------------------------------------------------------------
class ExponentialCEM(ChemicalEvolutionModel):
    r"""
    Exponentially declining star formation history model.

    This class models a galaxy's star formation rate as an exponentially
    declining function of time:

    .. math::
        M_\star(t) = M_{\rm inf} \cdot (1 - e^{-t/\tau})
 
    Attributes
    ----------
    stellar_mass_inf : astropy.Quantity
        Asymptotic stellar mass at infinite time.
    tau : astropy.Quantity
        Timescale of the exponential decline in star formation.
    metallicity : float
        Metallicity of the gas (constant).

    Examples
    --------
    Create an instance of the `ExponentialCEM` model and compute the stellar mass formed:

    >>> from astropy import units as u
    >>> from pst.models import ExponentialCEM
    >>> model = ExponentialCEM(stellar_mass_inf=1e10 * u.Msun, tau=2 * u.Gyr, metallicity=0.02)
    >>> time = [0.5, 1.0, 2.0] * u.Gyr
    >>> model.stellar_mass_formed(time)
    <Quantity [2.21199217e+09, 3.93469340e+09, 6.32120559e+09] solMass>    
    """
    def __init__(self, **kwargs):
        self.stellar_mass_inf = check_unit(kwargs['stellar_mass_inf'],
                                           default_unit=u.Msun)
        self.tau = check_unit(kwargs['tau'], default_unit=u.Gyr)
        self.metallicity = kwargs['metallicity']
        super().__init__(**kwargs)

    @u.quantity_input
    def stellar_mass_formed(self, time : u.Gyr):
      return self.stellar_mass_inf * ( 1 - np.exp(-time/self.tau) )

    @u.quantity_input
    def ism_metallicity(self, time : u.Gyr):
        return np.full(time.size, fill_value=self.metallicity)


class ExponentialQuenchedCEM(ExponentialCEM):
    """
    Exponentially declining CEM model including a quenching event.

    Examples
    --------
    Create an instance of the `ExponentialQuenchedCEM` model and compute the stellar mass formed:

    >>> from astropy import units as u
    >>> from pst.models import ExponentialQuenchedCEM
    >>> model = ExponentialQuenchedCEM(stellar_mass_inf=1e10 * u.Msun, tau=2 * u.Gyr,
    ...                                 metallicity=0.02, quenching_time=1.5 * u.Gyr)
    >>> time = [0.5, 1.0, 2.0] * u.Gyr
    >>> model.stellar_mass_formed(time)
    <Quantity [2.21199217e+09, 3.93469340e+09, 5.27633447e+09] solMass>

    See also
    --------
    :class:`ExponentialCEM`
    """
    def __init__(self, **kwargs):
        self.quenching_time = check_unit(kwargs['quenching_time'], u.Gyr)
        super().__init__(**kwargs)

    @sfh_quenching_decorator
    def stellar_mass_formed(self, time : u.Gyr):
        return super().stellar_mass_formed(time)


#-------------------------------------------------------------------------------
class ExponentialDelayedCEM(ChemicalEvolutionModel):
    r"""
    Exponentially delayed star formation history model.

    This CEM models a galaxy's star formation rate as a delayed exponential
    function of time, where the SFR rises initially and then decays.

    .. math::
        M_\star(t) = M_{\rm inf} \cdot (1 - \frac{t + \tau}{\tau} \cdot e^{-t/\tau})
 
    The main difference from the :class:`ExponentialCEM` is that the stellar mass 
    grows linearly at early times before transitioning to an exponential decay.

    Attributes
    ----------
    tau : astropy.Quantity
        Timescale of the delayed exponential star formation rate.
    today : float or astropy.Quantity
        Cosmic time at the the time of the observation.
    mass_today : float or astropy.Quantity
        Total stellar mass formed at present.
    ism_metallicity_today : float
        Metallicity of the gas (constant).
    
    Examples
    --------
    Create an instance of the `ExponentialDelayedCEM` model and compute the stellar mass formed:

    >>> from astropy import units as u
    >>> from pst.models import ExponentialDelayedCEM
    >>> model = ExponentialDelayedCEM(tau=2 * u.Gyr, today=10 * u.Gyr,
    ...                                mass_today=1e10 * u.Msun, ism_metallicity_today=0.02)
    >>> time = [0.5, 1.0, 2.0] * u.Gyr
    >>> model.stellar_mass_formed(time)
    <Quantity [2.76154498e+08, 9.40043900e+08, 2.75373844e+09] solMass>
    """

    def __init__(self, **kwargs):
        self.tau = check_unit(kwargs['tau'], default_unit=u.Gyr)
        self.today = check_unit(kwargs['today'], u.Gyr)
        self.mass_today = check_unit(kwargs['mass_today'], u.Msun)
        self._mass_norm = 1
        mtoday = self.stellar_mass_formed(self.today)
        self._mass_norm = self.mass_today / mtoday

        self.ism_metallicity_today =  kwargs['ism_metallicity_today']
        ChemicalEvolutionModel.__init__(self, **kwargs)

    @u.quantity_input
    def stellar_mass_formed(self, time):
        return self.mass_today * (1 - np.exp(-time / self.tau)
            * (self.tau + time) / self.tau) * self._mass_norm

    @u.quantity_input
    def ism_metallicity(self, time : u.Gyr):
        return np.full(time.size, fill_value=self.ism_metallicity_today)


class ExponentialDelayedZPowerLawCEM(MassPropMetallicityMixin, ExponentialDelayedCEM):
    """A :class:`ExponentialDelayedCEM` with a Mass-dependent Metallicity chemical model.
    
    See also
    --------
    :class:`MassPropMetallicityMixin`
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.ism_metallicity_today = kwargs["ism_metallicity_today"]
        self.alpha_powerlaw = kwargs["alpha_powerlaw"]


class ExponentialDelayedQuenchedCEM(ExponentialDelayedZPowerLawCEM):
    """A :class:`ExponentialDelayedZPowerLawCEM` with a quenching event."""
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.quenching_time = check_unit(kwargs["quenching_time"], u.Gyr)
        mtoday = self.stellar_mass_formed(self.today)
        self._mass_norm *= self.mass_today / mtoday

    @sfh_quenching_decorator
    def stellar_mass_formed(self, times: u.Quantity):
        return super().stellar_mass_formed(times)

#-------------------------------------------------------------------------------
class GaussianBurstCEM(ChemicalEvolutionModel):
    """
    Gaussian burst star formation model.

    This model represents a galaxy's star formation history as a single Gaussian 
    burst of star formation. The burst occurs at a specific time and is characterized 
    by a Gaussian distribution with a given standard deviation.

    Attributes
    ----------
    mass_burst : float or astropy.Quantity
        Total stellar mass formed in the burst.
    time_burst : float or astropy.Quantity
        Time of the starburst in cosmic time.
    sigma_burst : float or astropy.Quantity
        Span time of the burst in terms of the standard deviation.
    burst_metallicity : float or astropy.Quantity
        Metallicity of the burst.
    
    Examples
    --------
    Create an instance of the `GaussianBurstCEM` model and compute the stellar mass formed:

    >>> from astropy import units as u
    >>> from pst.models import GaussianBurstCEM
    >>> model = GaussianBurstCEM(mass_burst=1e10 * u.Msun, time_burst=2 * u.Gyr,
    ...                          sigma_burst=0.5 * u.Gyr, burst_metallicity=0.02)
    >>> time = [1.0, 2.0, 3.0] * u.Gyr
    >>> model.stellar_mass_formed(time)
    <Quantity [2.27501319e+08, 5.00000000e+09, 9.77249868e+09] solMass>
    """
 
    def __init__(self, **kwargs):
        self.mass_burst = check_unit(kwargs["mass_burst"], u.Msun)
        self.time_burst = check_unit(kwargs["time_burst"], u.Gyr)
        self.sigma_burst = check_unit(kwargs["sigma_burst"], u.Gyr)
        self.burst_metallicity = kwargs["burst_metallicity"]
        ChemicalEvolutionModel.__init__(self, **kwargs)

    @u.quantity_input
    def stellar_mass_formed(self, time):
        return self.mass_burst / 2 * (1 + special.erf(
            (time-self.time_burst) / (SQRT_2 * self.sigma_burst))
            )
  
    @u.quantity_input
    def ism_metallicity(self, time : u.Gyr):
        return np.full(time.size, fill_value=self.burst_metallicity)


class LogNormalCEM(ChemicalEvolutionModel):
    r"""
    Log-normal star formation history model.

    This CEM models a galaxy's star formation rate as a log-normal
    function of time (e.g. Gladders et al 2013), where the SFR rises initially
    and then decays:

    .. math::
        M_\star(t) = \frac{M_{\rm today}}{2} \cdot \left(1 - {\rm erf}\left(\frac{{\rm ln}(t) - {\rm ln}(t_0)}{\sigma \sqrt{2}}\right) \right)

    Attributes
    ----------
    lnt0 : float
        Mean of the lognormal SFH.
    tau : astropy.Quantity
        Standard deviation of the lognormal SFH.
    mass_today : float or astropy.Quantity
        Total stellar mass formed at present.
    metallicity : float
        Metallicity of the gas (constant).
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.t0 = check_unit(kwargs['t0'], u.Gyr)
        self.scale = kwargs['scale']
        self.today = check_unit(kwargs['today'], u.Gyr)
        self.mass_today = check_unit(kwargs['mass_today'], u.Msun)
        self.mass_norm = 1
        mtoday = self.stellar_mass_formed(self.today)
        self.mass_norm = self.mass_today / mtoday
        self.ism_metallicity_today =  kwargs['ism_metallicity_today']

    @u.quantity_input
    def stellar_mass_formed(self, times: u.Quantity):
        z = - np.log(times[times > 0] / self.t0) / self.scale
        m = np.zeros(times.shape)
        m[times > 0] = 0.5 * (1 - special.erf(z / SQRT_2))
        return m * self.mass_norm

    @u.quantity_input
    def ism_metallicity(self, time : u.Gyr):
        return np.full(time.size, fill_value=self.ism_metallicity_today)


class LogNormalZPowerLawCEM(MassPropMetallicityMixin, LogNormalCEM):
    """A :class:`LogNormalCEM` with a Mass-dependent Metallicity chemical model.
    
    See also
    --------
    :class:`MassPropMetallicityMixin`
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.ism_metallicity_today = kwargs["ism_metallicity_today"]
        self.alpha_powerlaw = kwargs["alpha_powerlaw"]


class LogNormalQuenchedCEM(LogNormalZPowerLawCEM):
    """A :class:`LogNormalCEM` with a quenching event."""
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.quenching_time = check_unit(kwargs["quenching_time"], u.Gyr)
        mtoday = self.stellar_mass_formed(self.today)
        self.mass_norm *= self.mass_today / mtoday

    @sfh_quenching_decorator
    def stellar_mass_formed(self, times: u.Quantity):
        return super().stellar_mass_formed(times)


#-------------------------------------------------------------------------------
class TabularCEM(ChemicalEvolutionModel):
    """Chemical evolution model based on a grid of times and metallicities.
    
    Description
    -----------
    This model represents the chemical evolution of a galaxy by means of a
    discrete grid of ages and metallicities

    Attributes
    ----------
    table_t: astropy.Quantity
        Tabulated cosmic time.
    table_M: astropy.Quantity
        Total stellar mass at each cosmic time step.
    table_metallicity: astropy.Quantity
        ISM metallicity at each cosmic time step.

    See also
    --------
    :class:`pst.models.ChemicalEvolutionModel` documentation.
    """
    def __init__(self, times, masses, metallicities, **kwargs):
        super().__init__(**kwargs)
        self.table_t = check_unit(times, u.Gyr)
        # Make sure that time is crescent
        sort_times = np.argsort(self.table_t)
        self.table_t = self.table_t[sort_times]
        self.table_mass = check_unit(masses[sort_times], u.Msun)
        self.table_metallicity = metallicities[sort_times]

    @u.quantity_input
    def stellar_mass_formed(self, times: u.Gyr):
        r"""Evaluate the integral of the SFR over a given set of times.
        
        Description
        -----------
        This method evaluates the integral:

        .. math::
            \int_{0}^{t} {\rm SFR}(t') dt'

        at each time input time :math:`t`.

        Parameters
        ----------
        times : :class:`astropy.units.Quantity`
            Array of cosmic times at which the integral will be evaluated.

        Returns
        -------
        integral : :class:`astropy.units.Quantity`
            The cumulative stellar mass formed at each input time.
        """
        interpolator = interpolate.PchipInterpolator(
           self.table_t, self.table_mass)
        integral = interpolator(times.to_value(self.table_t.unit)
                                ) << self.table_mass.unit
        integral[times > self.table_t[-1]] = self.table_mass[-1]
        integral[times < self.table_t[0]] = 0
        return integral

    @u.quantity_input
    def ism_metallicity(self, times: u.Gyr):
        """Evaluate the integral of the SFR over a given set of times.
        
        Description
        -----------
        Return the metallicity Z(t) of the interstellar medium (gas and dust)
        at a certain set of cosmic times (i.e. since the Big Bang).
        
        Parameters
        ----------
        times: astropy.units.Quantity
            Cosmic times at which the metallicity will be evaluated.

        Returns
        -------
        z_t: astropy.units.Quantity
            Vector with the ISM metallicity at each input time.
        """
        interpolator = interpolate.PchipInterpolator(
           self.table_t, self.table_metallicity)
        integral = interpolator(times)
        integral[times > self.table_t[-1]] = self.table_metallicity[-1]
        integral[times < self.table_t[0]] = self.table_metallicity[0]
        return integral


class TabularCEM_ZPowerLaw(MassPropMetallicityMixin, TabularCEM):
    """Chemical evolution model based on a grid of times that assumes an analytic chemical enrichment history.

    See Also
    --------
    :class:`TabularCEM`
    :class:`MassPropMetallicityMixin`

    """
    def __init__(self, times, masses, alpha_powerlaw, ism_metallicity_today,
                 mass_today, **kwargs):
        self.ism_metallicity_today = ism_metallicity_today
        self.alpha_powerlaw = alpha_powerlaw
        self.mass_today = mass_today
        # Create a dummy metallicity that is passed to the TabularCEM constructor
        # but never used
        metallicity = np.zeros(times.size)
        super().__init__(times, masses, metallicity, **kwargs)


class ParticleListCEM(ChemicalEvolutionModel):
    """
    Chemical Evolution Model using individual Simple Stellar Population (SSP) data.
    
    This model represents the chemical evolution of a galaxy by reconstructing 
    a composite stellar population (CSP) using individual stellar population (SSP) 
    particles, each of which is defined by its formation time, metallicity, and mass.

    Parameters
    ----------
    time_form : numpy.array or astropy.units.Quantity
        Array representing the formation times of each SSP particle. If the input 
        is a `numpy.array`, it is assumed to be in Gyr. Otherwise, an `astropy.units.Quantity` 
        with appropriate units is required.
    metallicities : numpy.array or astropy.units.Quantity
        Array representing the metallicities of each SSP particle. If the input 
        is a `numpy.array`, it is assumed to be dimensionless (i.e., no units).
    masses : numpy.array or astropy.units.Quantity
        Array representing the masses of each SSP particle. If the input is a 
        `numpy.array`, it is assumed to be in solar masses.
    """
    def __init__(self, time_form, metallicities, masses):
        self.time_form, self.metallicities, self.masses = (
            time_form, metallicities, masses)

    @property
    def time_form(self) -> u.Quantity:
        """Array of SSP formation times in Gyr."""
        return self._time_form

    @time_form.setter
    def time_form(self, values):
        if not isinstance(values, u.Quantity):
            self._time_form = values << u.Gyr
        else:
            self._time_form = values

    @property
    def metallicities(self) -> u.Quantity:
        """Array of SSP metallicities, assumed to be dimensionless."""
        return self._metallicities

    @metallicities.setter
    def metallicities(self, values):
        if not isinstance(values, u.Quantity):
            self._metallicities = values << u.dimensionless_unscaled
        else:
            self._metallicities = values

    @property
    def masses(self) -> u.Quantity:
        """Array of SSP masses in solar masses."""
        return self._masses

    @masses.setter
    def masses(self, values):
        if not isinstance(values, u.Quantity):
            self._masses = values << u.Msun
        else:
            self._masses = values

    def interpolate_ssp_masses(self, ssp: SSPBase, t_obs: u.Quantity):
        """
        Interpolate the SSP particles onto an SSP base model at the observed time.
        
        This method interpolates the stellar masses of the SSP particles based on 
        their ages and metallicities at the observed cosmic time `t_obs` using an 
        SSP model grid.

        Parameters
        ----------
        ssp : pst.SSP.SSPBase
            SSP model providing the ages and metallicities for interpolation.
        t_obs : astropy.units.Quantity
            The age of the Universe at the time of the observation.

        Returns
        -------
        ssp_weights : astropy.units.Quantity
            A 2D array representing the stellar mass associated with each SSP 
            particle in the base SSP grid. Units are in solar masses.
        """
        valid_particles = self.time_form <= t_obs
        return ssp.get_weights(ages=t_obs - self.time_form[valid_particles],
                               metallicities=self.metallicities[valid_particles],
                               masses=self.masses[valid_particles])

    def stellar_mass_formed(self, time):
        sort_idx = np.argsort(self.time_form)
        mass_history = np.cumsum(self.masses[sort_idx])
        return np.interp(time, self.time_form[sort_idx], mass_history)

    def ism_metallicity(self, time):
        return np.full(time.size, fill_value=np.nan)

# -----------------------------------------------------------------------------
#                                                    ... Paranoy@ Rulz! ;^D
# Mr Krtxo \(ﾟ▽ﾟ)/
# -----------------------------------------------------------------------------



"""
This module contains some tools for computing observable quantities
(e.g. photometry, equivalent widths) from spectra.
"""

import numpy as np
import os
from astropy import units as u
from astropy import constants
import requests
import json
from matplotlib import pyplot as plt
from . import utils

PST_DATA_DIR = os.path.join(os.path.dirname(__file__), "data")

def list_of_available_filters():
    """List the currently available filters in the default directory."""
    filter_dir = os.path.join(PST_DATA_DIR, "filters")
    return os.listdir(filter_dir)

def load_photometric_filters(filters):
    """Convenience function for constructing a list of photometric filters.
    
    Parameters
    ----------
    filters: list of str
        List of filters to load. The list might contain the absolute path to
        a filter response file, or just the filter name following the SVO
        convention.
    
    Returns
    -------
    filters_out: list of :class:`pst.observables.Filter`
        List of filters.
    """
    filters_out = []
    for f in filters:
        if os.path.exists(f):
            filters_out.append(Filter.from_text_file(f))
        else:
            filters_out.append(Filter.from_svo(f))
    return filters_out

def download_svo_filter(name: str, dest_dir: str, verbose=True):
    """Download a filter from the Spanish Virtual Observatory (SVO) Filter Profile Service.
    
    Parameters
    ----------
    name : str
        SVO-compliant filename. The naming convention for SVO filters is 
        TELESC_INSTRUMENT.BAND (e.g. WISE_WISE.W1, Subaru_HSC.g)
    dest_dir : str
        Path to the directory where to store the data.
    
    Returns
    -------
    file_path : str
        Path to the downloaded filter file.
    """
    if not os.path.exists(dest_dir):
        os.makedirs(dest_dir)  # create folder if it does not exist
    name = name.strip(".dat")
    base_url="http://svo2.cab.inta-csic.es/theory/fps/getdata.php?format=ascii&id="
    url = base_url + name.replace("_", "/")
    filename = name + ".dat"
    file_path = os.path.join(dest_dir, filename)
    if verbose:
        print(f"Querying SVO Filter: {url}")
    r = requests.get(url, stream=True)
    if len(r.text) > 0:
        if verbose:
            print(f"Saving new filter {name} to ", os.path.abspath(file_path))
        with open(file_path, 'wb') as f:
            for chunk in r.iter_content(chunk_size=1024 * 8):
                if chunk:
                    f.write(chunk)
                    f.flush()
                    os.fsync(f.fileno())
        return file_path
    else:
        raise FileNotFoundError("Query to {url} was unsucessful")


class Filter(object):
    """A photometric filter.
    
    Attributes
    ----------
    filter_resp: np.ndarray
        Original photometric passband response curve.
    filter_wavelength : :class:`numpy.ndarray` or :class:`astropy.units.Quantity`
        Original wavelength associated to ``filter_resp``. If type is
        ``numpy.ndarray``, the value is converted in to an ``astropy.units.Quantity`
        in Angstrom.
    response: np.ndarray
        Filter passband response curve after interpolation.
    wavelength: :class:`numpy.ndarray` or :class:`astropy.units.Quantity`
        Wavelength vector associated to ``response``. If type is
        ``numpy.ndarray``, the value is converted in to an ``astropy.units.Quantity`
        in Angstrom.
    default_dir : str
        Default directory containing filter files.

    Example
    -------
    >>> from pst.observables import Filter
    >>> ps_r_filter = Filter("PANSTARRS_PS1.r")
    """

    default_dir = os.path.join(PST_DATA_DIR, "filters")

    def __init__(self, wavelength=None, response=None,
                 filter_wavelength=None, filter_response=None):

        self.wavelength = wavelength
        self.response = response
        self.filter_wavelength = filter_wavelength
        self.filter_resp = filter_response
        if filter_wavelength is None and filter_response is None and wavelength is not None:
            self.filter_wavelength = self.wavelength
            self.filter_resp = self.response
        elif filter_wavelength is None and wavelength is None:
            raise NameError("wavelength or filter_wavelength must be provided")

    @property
    def wavelength(self):
        return self._wavelength

    @wavelength.setter
    def wavelength(self, value):
        if not isinstance(value, u.Quantity) and value is not None:
            self._wavelength = value << u.angstrom
        else:
            self._wavelength = value

    @property
    def filter_wavelength(self):
        return self._filter_wavelength

    @filter_wavelength.setter
    def filter_wavelength(self, value):
        if not isinstance(value, u.Quantity) and value is not None:
            self._filter_wavelength = value << u.angstrom
        else:
            self._filter_wavelength = value

    @classmethod
    def from_text_file(cls, path, wavelength_unit=u.angstrom, **kwargs):
        """Load a :class:``Filter`` from an input text file.

        Parameters
        ---------
        path : str
            Path to the text file containing the filter information.
            The first and second columns must correspond to the
            wavelength and passband curve, respectively.
        wavelength_units : :class:``astropy.units.Quantity``
        **kwargs : 
            Arguments to be passed to :func:`numpy.loadtxt`

        Returns
        -------
        filter : :class:`Filter`
            The ``Filter`` containing the input information.
        """
        wavelength, response = np.loadtxt(path, usecols=(0, 1), unpack=True,
                                          **kwargs)
        return cls(filter_wavelength=wavelength * wavelength_unit,
                   filter_response=response)

    @classmethod
    def from_svo(cls, name, destination_dir=None):
        """Load a :class:``Filter`` from the Spanish Vitural Observatory archive.

        Parameters
        ---------
        name : str
            SVO filter name. If the filter is not found locally, it will be
            downloaded from the archive.

        Returns
        -------
        filter : :class:`Filter`
            The ``Filter`` containing the input information.

        Example
        -------
        >>> from pst.observables import Filter
        >>> panstarrs_r_filter = Filter.from_svo("PANSTARRS_PS1.r")
        """
        if destination_dir is None:
            destination_dir = cls.default_dir
        path = cls._isfilter(name)
        if path is not None:
            return cls.from_text_file(path)
        else:
            path = download_svo_filter(name, dest_dir=destination_dir)
            return cls.from_text_file(path)

    @classmethod
    def _isfilter(cls, name):
        path = os.path.join(cls.default_dir, name.strip(".dat") + ".dat")
        if os.path.isfile(path):
            return path
        else:
            return None

    def effective_wavelength(self):
        r"""Compute the effective wavelength of the filter.
        
        Description
        -----------
        The effective wavelength is computed as

        .. math::
            \lambda_{\rm eff} = \frac{\int{R(\lambda) \cdot \lambda d\lambda}}{\int{R(\lambda) d\lambda}}
        
        Returns
        -------
        eff_wl : :class:`astropy.units.Quantity`
            The effective wavelength of the filter.
        """
        return np.sum(self.filter_wavelength*self.filter_resp)/np.sum(self.filter_resp)

    def effective_bandwidth(self):
        r"""Compute the effective bandwidth of the filter.
        
        Description
        -----------
        The effective bandwith is computed as

        .. math::
            \Delta \lambda_{\rm BW} = \sqrt{8\log(2)} \left(\frac{\int{R(\lambda) \cdot \lambda^2 d\lambda}}{\int{R(\lambda) d\lambda}} - \lambda_{\rm eff}\right)^{1/2}

        Returns
        -------
        eff_bw : :class:`astropy.units.Quantity`
            The effective bandwidth of the filter.
        
        See also
        --------
        :func:`effective_wavelength`
        """
        return np.sqrt(8*np.log(2)*(
            np.sum(self.filter_wavelength**2*self.filter_resp)/np.sum(self.filter_resp)
            - self.effective_wavelength()**2))

    def effective_transmission(self):
        r"""Compute the effective bandwidth of the filter.
        
        Description
        -----------
        The effective transmission is computed as

        .. math::
            R_{\rm eff} = \frac{\int{R(\lambda)^2 d\lambda}}{\int{R(\lambda) d\lambda}}

        Returns
        -------
        eff_tr : float
            The effective transmission of the filter.
        """
        return np.sum(self.filter_resp**2)/np.sum(self.filter_resp)

    def interpolate(self, wavelength=None):
        """Interpolate and update the Filter response curve to an input wavelength.
        
        Description
        -----------
        Interpolate linearly the Filter response curve to an input wavelength vector.
        The result will update the exising values of ``wavelength`` and ``response``.

        Parameters
        ----------
        wavelength: :class:`numpy.ndarray` or :class:`astropy.units.Quantity`
            Wavelength vector to interpolate ``filt_resp``. If type is
            ``numpy.ndarray``, the value is converted in to an ``astropy.units.Quantity`
            in Angstrom.
        
        Returns
        -------
        response : np.ndarray
            Filter response curve interpolated to the input values of ``wavelength``.
        """
        if not hasattr(wavelength, "unit"):
            wavelength = wavelength << u.angstrom

        self.response = utils.flux_conserving_interpolation(
            wavelength, self.filter_wavelength, self.filter_resp)
        self.wavelength= wavelength
        return self.response

    def get_photons(self, spectra, spectra_err=None, mask_nan=True):
        r"""Compute the photon flux from an input spectra.
        
        Description
        -----------
        The photon flux associated to the filter is computed by numerically integrating
        the input ``spectra`` with the filter ``response``, using the trapezid method.

        .. :math:
            phot_flux = \int{F_\lambda \cdot frac{\lambda}{hc} \cdot R(\lambda) d\lambda}

        Parameters
        ----------
        spectra : :class:`np.ndarray` or :class:``astropy.units.Quantity``
            Input spectra (flux density per wavelength unit) with same
            dimensions as the Filter ``wavelength``.
        spectra_err : :class:`np.ndarray` or :class:``astropy.units.Quantity``, optional
            Input spectra associated error.
        mask_nan : bool, optional
            If True, NaN values are masked.
        
        Returns
        -------
        photon_flux : :class:``astropy.units.Quantity``
            Filter photon flux.
        photon_flux_err : :class:``astropy.units.Quantity``
            Filter photon flux associated error.
        """
        spectra = utils.check_unit(spectra, default_unit=u.Lsun / u.angstrom / u.cm**2,
                                   equivalence=u.spectral_density, wav=self.wavelength)

        if mask_nan:
            mask = np.isfinite(spectra)
            photon_flux = np.trapz(
                spectra[mask] / (constants.h * constants.c / self.wavelength[mask]
                                   ) * self.response[mask],
                x=self.wavelength[mask])
        else:
            photon_flux = np.trapz(
                spectra / (constants.h * constants.c / self.wavelength
                                   ) * self.response,
                x=self.wavelength)

        if spectra_err is not None:

            spectra_err = utils.check_unit(spectra_err,
                                           default_unit=u.Lsun / u.angstrom / u.cm**2,
                                           equivalence=u.spectral_density, wav=self.wavelength)
            if mask_nan:
                mask = mask & np.isfinite(spectra_err)
            else:
                mask = np.ones_like(spectra_err, dtype=bool)

            photon_flux_err = np.trapz(
                spectra_err[mask] / (constants.h * constants.c / self.wavelength[mask]
                                       ) * self.response[mask],
                x=self.wavelength[mask])
        else:
            photon_flux_err = None
        return photon_flux, photon_flux_err
    
    def get_ab(self, spectra, spectra_err=None, mask_nan=True):
        r"""Compute the synthetic AB magnitude from an input spectra.
        
        Description
        -----------
        The AB magnitude associated to the filter is computed by numerically integrating
        the input ``spectra`` with the filter ``response``, using the trapezid method.

        .. :math:
            phot_flux = -2.5 \cdot \log_{10}\left(\frac{N_{phot}(spectra)}{N_{phot}(3631)}\right)

        Parameters
        ----------
        spectra : :class:`np.ndarray` or :class:``astropy.units.Quantity``
            Input spectra (flux density per wavelength unit) with same
            dimensions as the Filter ``wavelength``.
        spectra_err : :class:`np.ndarray` or :class:``astropy.units.Quantity``, optional
            Input spectra associated error.
        mask_nan : bool, optional
            If True, NaN values are masked.
        
        Returns
        -------
        mag_ab : :class:``astropy.units.Quantity``
            AB magnitude.
        mag_ab_err : :class:``astropy.units.Quantity``
            AB magnitude associated error.
        
        See also
        --------
        :func:`get_photons`
        """
        n_photons, n_photons_err = self.get_photons(spectra, spectra_err, mask_nan=mask_nan)
        norm_photons, _ = self.get_photons(
            3630.781 * u.Jy * np.ones(spectra.shape) * constants.c / self.wavelength**2,
            mask_nan=False)
        mag_ab = - 2.5 * np.log10(n_photons / norm_photons)
        if n_photons_err is None:
            mag_ab_err = None
        else:
            mag_ab_err = 2.5 / np.log(10) * n_photons_err / n_photons
        return mag_ab, mag_ab_err

    def get_fnu(self, spectra, spectra_err=None, mask_nan=True):
        """Compute the  specific flux per frequency unit from a spectra.

        Parameters
        ----------
        spectra : :class:`np.ndarray` or :class:``astropy.units.Quantity``
            Input spectra (flux density per wavelength unit) with same
            dimensions as the Filter ``wavelength``.
        spectra_err : :class:`np.ndarray` or :class:``astropy.units.Quantity``, optional
            Input spectra associated error.
        mask_nan : bool, optional
            If True, NaN values are masked.

        Returns
        -------
        mag_ab : :class:``astropy.units.Quantity``
            AB magnitude.
        mag_ab_err : :class:``astropy.units.Quantity``
            AB magnitude associated error.

        See also
        --------
        :func:`get_photons`
        """
        n_photons, n_photons_err = self.get_photons(spectra, spectra_err, mask_nan=mask_nan)
        norm_photons, _ = self.get_photons(
            3630.781 * u.Jy * np.ones(spectra.shape) * constants.c / self.wavelength**2,
            mask_nan=False)
        f_nu = n_photons / norm_photons * 3630.781 * u.Jy
        if spectra_err is None:
            f_nu_err = None
        else:
            f_nu_err = n_photons_err / norm_photons * 3630.781 * u.Jy
        return f_nu, f_nu_err

    def get_flambda_vegamag(self, spectra, spectra_err=None, mask_nan=True):
        """Compute the  specific flux per wavelength unit from a spectra.

        Parameters
        ----------
        spectra : :class:`np.ndarray` or :class:``astropy.units.Quantity``
            Input spectra (flux density per wavelength unit) with same
            dimensions as the Filter ``wavelength``.
        spectra_err : :class:`np.ndarray` or :class:``astropy.units.Quantity``, optional
            Input spectra associated error.
        mask_nan : bool, optional
            If True, NaN values are masked.

        Returns
        -------
        mag_ab : :class:``astropy.units.Quantity``
            AB magnitude.
        mag_ab_err : :class:``astropy.units.Quantity``
            AB magnitude associated error.

        See also
        --------
        :func:`get_photons`
        """
        spectra = spectra = utils.check_unit(spectra, default_unit=u.Lsun / u.angstrom / u.cm**2,
                                   equivalence=u.spectral_density, wav=self.wavelength)

        if mask_nan:
            mask = np.isfinite(spectra)
        else:
            mask = np.ones_like(spectra, dtype=bool)

        f_lambda = np.trapz(spectra[mask] * self.wavelength[mask] * self.response[mask], x=self.wavelength[mask]
                               ) / np.trapz(self.response[mask] * self.wavelength[mask], x=self.wavelength[mask])

        if spectra_err is not None:

            spectra_err = utils.check_unit(spectra_err, default_unit=u.Lsun / u.angstrom / u.cm**2,
                                           equivalence=u.spectral_density, wav=self.wavelength)
            if mask_nan:
                mask = mask & np.isfinite(spectra_err)
            else:
                mask = np.ones_like(spectra_err, dtype=bool)

            f_lambda_err = np.trapz(
            spectra_err[mask] / (constants.h * constants.c / self.wavelength[mask]
                                   ) * self.response[mask],
            x=self.wavelength[mask])
        else:
            f_lambda_err = None

        return f_lambda, f_lambda_err

    def plot(self, show=False):
        """Plot the filter response curve.
        
        Plot the original filter response curve together with the interpolated
        version computed using a new grid of wavelengths.
        """
        fig = plt.figure()
        ax = fig.add_subplot(111)
        ax.step(self.filter_wavelength, self.filter_resp, label='Original',
                color='k', where="mid")
        ax.plot(self.filter_wavelength, self.filter_resp, '.', color='k')
        ax.set_xlabel(f"Wavelength ({self.filter_wavelength.unit})")
        ax.set_ylabel("Filter response")
        if self.wavelength is not None:
            ax.step(self.wavelength, self.response, label='Interpolated',
                      color='r', where="mid")
        ax.legend()
        if show:
            plt.show()
        else:
            plt.close()
        return fig


class TopHatFilter(Filter):
    """Top hat photometric filter
    
    See also
    --------
    :class:`Filter`
    """
    def __init__(self, central_wave, width, **kwargs):
        central_wave = utils.check_unit(central_wave, u.Angstrom)
        width = utils.check_unit(width, u.Angstrom)

        self.wavelength = kwargs.get('wavelength', None)
        if self.wavelength is None:
            self.filter_wavelength = np.linspace(central_wave - width,
                                    central_wave + width,
                                    50)
        else:
            self.wavelength = utils.check_unit(self.wavelength, u.Angstrom)
            self.filter_wavelength = self.wavelength.copy()

        self.filter_resp = np.ones(self.filter_wavelength.size)
        self.filter_resp[self.filter_wavelength < central_wave - width / 2] = 0
        self.filter_resp[self.filter_wavelength > central_wave + width / 2] = 0
        if self.wavelength is None:
            self.response = self.filter_resp.copy()


class EquivalentWidth(object):
    r"""Equivalent width of an spectral region.
    
    Description
    -----------
    Given a stellar spectra spectra :math:`F_\lambda`, the equivalent width is
    defined as the area of the spectral line, defined over a given spectral
    region (:math:`\lambda_{\rm C,\,min}, \lambda_{\rm C,\,max}`), divided by
    the average flux of the continuum in a given spectral region.
    It is computed as: 

    .. math::
        EW = \int_{\lambda_{C,\,\rm min}}^{\lambda_{\rm C,\,max}} \left(1 - \frac{F_\lambda(\lambda)}{F_{\rm cont}(\lambda)}\right) d\lambda
    where:

    .. math::
        F_{cont}(\lambda) = \frac{F_{\rm B} \lambda_{\rm R} - F_{\rm R} \lambda_{\rm B}}{\lambda_{\rm R} - \lambda_{\rm B}} + \lambda \frac{F_{\rm R} - F_{\rm B}}{\lambda_{\rm R} - \lambda_{\rm B}}
        
    and :math:`F_{\rm B}` and :math:`F_{\rm R}` are the average flux in the left
    and right spectral windows, respectively, defined as:

    .. math::
        F_{\rm B} = \frac{1}{\lambda_{\rm B,\,max} - \lambda_{\rm B,\,min}} \int_{\lambda_{\rm B,\,min}}^{\lambda_{\rm B,\,max}} F_\lambda(\lambda) d\lambda
    
    .. math::
        F_{\rm R} = \frac{1}{\lambda_{\rm R,\,max} - \lambda_{\rm R,\,min}} \int_{\lambda_{\rm R,\,min}}^{\lambda_{\rm R,\,max}} F_\lambda(\lambda) d\lambda

    where :math:`\lambda_{\rm B,\,min}` and :math:`\lambda_{\rm B,\,max}` are the
    left spectral window boundaries, and :math:`\lambda_{\rm R,\,min}` and
    :math:`\lambda_{\rm R,\,max}` are the right spectral window boundaries.

    Example
    -------
    >>> from pst.observables import EquivalentWidth
    >>> ew = EquivalentWidth(left_wl_range=(3700, 3900),
    ...                      central_wl_range=(4000, 4100),
    ...                      right_wl_range=(4200, 4400))
    >>> wavelength = np.linspace(3600, 4600, 1000) * u.angstrom
    >>> spectra = np.random.normal(1, 0.1, size=wavelength.size) * u.erg / u.s / u.cm**2 / u.angstrom
    >>> ew_value, ew_err = ew.compute_ew(wavelength, spectra)
    """
    def __init__(self, left_wl_range, central_wl_range, right_wl_range):
        self.left_wl_range = np.array(left_wl_range)
        self.central_wl_range = np.array(central_wl_range)
        self.right_wl_range = np.array(right_wl_range)
    
    @property
    def left_wl_range(self) -> u.Quantity:
        r"""Spectral range defining the left pseudocontinuum window :math:`\lambda_{\rm B,\,min}, \lambda_{\rm B,\,max}`."""
        return self._left_wl_range
    
    @left_wl_range.setter
    def left_wl_range(self, value):
        if not isinstance(value, u.Quantity):
            self._left_wl_range = value * u.angstrom
        else:
            self._left_wl_range = value

    @property
    def right_wl_range(self) -> u.Quantity:
        r"""Spectral range defining the right pseudocontinuum window :math:`\lambda_{\rm R,\,min}, \lambda_{\rm R,\,max}`."""
        return self._right_wl_range

    @right_wl_range.setter
    def right_wl_range(self, value):
        if not isinstance(value, u.Quantity):
            self._right_wl_range = value * u.angstrom
        else:
            self._right_wl_range = value
    
    @property
    def central_wl_range(self) -> u.Quantity:
        r"""Spectral range defining the equivalent width window :math:`\lambda_{\rm C,\,min}, \lambda_{\rm C,\,max}`."""
        return self._central_wl_range

    @central_wl_range.setter
    def central_wl_range(self, value):
        if not isinstance(value, u.Quantity):
            self._central_wl_range = value * u.angstrom
        else:
            self._central_wl_range = value

    def compute_ew(self, wavelength, spectra, spectra_err=None):
        """Compute the equivalent width of a given input spectra.
        
        Description
        -----------
        The equivalent width is computed using the definition given in the class
        description. Positive values of the equivalent width indicate an absorption
        line, while negative values indicate an emission line. The error on the
        equivalent width is computed using the error propagation formula, assuming
        null covariance between the spectral points.

        Parameters
        ----------
        spectra : :class:`np.ndarray` or :class:``astropy.units.Quantity``
            Input spectra. If the array is multidimensional, the first axis must
            correspond to the spectral direction.
        wavelength : :class:`np.ndarray` or :class:``astropy.units.Quantity``
            Wavelength array associated to ``spectra``.
        spectra_err : :class:`np.ndarray` or :class:``astropy.units.Quantity``, optional
            If provided, computed the associated error of the equivalent width.
        
        Returns
        -------
        ew : np.ndarray
            The equivalent width of the input spectra.
        ew_err : np.ndarray
            The associated error of the equivalent width.
        """

        left_pts = np.where(np.searchsorted(self.left_wl_range, wavelength) == 1)[0]
        right_pts = np.where(np.searchsorted(self.right_wl_range, wavelength) == 1)[0]
        lick_pts = np.where(np.searchsorted(self.central_wl_range, wavelength) == 1)[0]
        delta_ew = self.central_wl_range[1] - self.central_wl_range[0]
        # pseudo-continuum interpolation
        right_weight = (self.central_wl_range.mean() - self.left_wl_range.mean()
                        ) / (self.right_wl_range.mean() - self.left_wl_range.mean())

        if spectra.ndim > 1:
            left_cont = np.nanmean(spectra[left_pts, :], axis=0)
            right_cont = np.nanmean(spectra[right_pts, :], axis=0)
            pseudocont = (1 - right_weight) * left_cont + right_weight * right_cont
            flux = np.nanmean(spectra[lick_pts], axis=0)
            ew = delta_ew * (1 - flux/pseudocont)
            if spectra_err is None:
                ew_err = np.nan
            else:
                left_cont_var = np.nanmean(spectra_err[left_pts, :]**2,
                                        axis=0) / left_pts.size
                right_cont_var = np.nanmean(spectra_err[right_pts, :]**2,
                                            axis=0) / right_pts.size
                pseudocont_var = ((1 - right_weight) * left_cont_var
                                + right_weight * right_cont_var)
                flux_var = np.nanmean(spectra_err[lick_pts]**2, axis=0)
                ew_var = ((delta_ew / pseudocont)**2 * flux_var
                        + (delta_ew * flux/pseudocont**2)**2 * pseudocont_var)
                ew_err = np.sqrt(ew_var)
        else:
            left_cont = np.nanmean(spectra[left_pts])
            right_cont = np.nanmean(spectra[right_pts])
            pseudocont = (1 - right_weight) * left_cont + right_weight * right_cont
            flux = np.nanmean(spectra[lick_pts])
            ew = delta_ew * (1 - flux/pseudocont)

            if spectra_err is None:
                ew_err = np.nan
            else:
                left_cont_var = np.nanmean(spectra_err[left_pts]**2
                                        ) / left_pts.size
                right_cont_var = np.nanmean(spectra_err[right_pts]**2
                                            ) / right_pts.size
                pseudocont_var = ((1 - right_weight) * left_cont_var
                                + right_weight * right_cont_var)
                flux_var = np.nanmean(spectra_err[lick_pts]**2)
                ew_var = delta_ew**2 * (
                    flux_var / lick_pts.size / pseudocont**2
                    + flux**2 * pseudocont_var / pseudocont**4)
                ew_err = np.sqrt(ew_var)
        return ew, ew_err

    @classmethod
    def from_json(cls, path):
        """Load a :class:`EquivalentWidth` from a JSON file.
        
        Parameters
        ----------
        path : str
            Path to the JSON file.
        
        Returns
        -------
        ew : :class:`EquivvalentWidth`
        """
        with open(path, "r") as f:
            data = json.load(f)
        return cls(**data)
    
    @classmethod
    def from_name(cls, name):
        """Load a :class:`EquivalentWidth` from a JSON file.
        
        Parameters
        ----------
        name : str
            Name of the Lick index.

        Returns
        -------
        ew : :class:`EquivvalentWidth`
        """
        json_file = os.path.join(PST_DATA_DIR, "lick", name + ".json")
        if os.path.isfile(json_file):
            return cls.from_json(json_file)
        else:
            raise FileNotFoundError(f"There is no JSON file\n -{json_file}"
                                    f"associated to input name {name}")



import unittest
import os
import numpy as np
from astropy import units as u
from astropy import constants
from pst import observables

class TestObservables(unittest.TestCase):

    @classmethod
    def setUpClass(self):
        print("Setting SSP model")
        self.dummy_wavelength = np.geomspace(100, 1e5, 3000) * u.angstrom
        # Monocromatic SED
        self.dummy_flam = np.ones(self.dummy_wavelength.size
                                     ) * constants.c / self.dummy_wavelength**2 * 3631 * u.Jy
        self.dummy_fnu = np.ones(self.dummy_wavelength.size
                                ) * 3631 * u.Jy

    def test_default_dir(self):
        self.assertTrue(
            os.path.isdir(observables.PST_DATA_DIR),
            f"Default PST data dir {observables.PST_DATA_DIR} does not exists")

    def test_list_filters_available(self):
        paths = observables.list_of_available_filters()
        self.assertIsNotNone(paths)

    def test_load_photometric_filters(self):
        filters = observables.load_photometric_filters(
            ["SLOAN_SDSS.u", "SLOAN_SDSS.g", "SLOAN_SDSS.r"])
        self.assertIsNotNone(filters)
        for f in filters:
            self.assertTrue(isinstance(f, observables.Filter))
    
    def test_filter(self):
        filter = observables.Filter.from_svo("SLOAN_SDSS.i")
        self.assertIsNotNone(filter)
        # Reload filter from file
        filter = observables.Filter.from_text_file(
            os.path.join(observables.PST_DATA_DIR, "filters", "SLOAN_SDSS.i.dat")
        )
        self.assertIsNotNone(filter)

        self.assertTrue(np.isclose(filter.effective_wavelength(),
                                   7499.70417446 * u.angstrom),
                        "Unexpected effective wavelength value")
        
        self.assertTrue(np.isclose(filter.effective_bandwidth(),
                                   902.02184282 * u.angstrom),
                        "Unexpected effective bandwidth value")
        
        self.assertTrue(np.isclose(filter.effective_transmission(),
                                   0.32484839450189695),
                        "Unexpected effective transmission value")

        # Interpolate filter to input wavelength array
        filter.interpolate(self.dummy_wavelength)
        # Use flam
        flux, _ = filter.get_fnu(self.dummy_flam)
        self.assertTrue(np.isclose(flux, 3631.0 * u.Jy),
                        f"Unexpected integrated flux value: {flux}")
        
        mag, _ = filter.get_ab(self.dummy_flam)
        self.assertTrue(np.isclose(mag, 0.0, atol=1e-4),
                        f"Unexpected magnitude value: {mag}")
        # Use fnu
        flux, _ = filter.get_fnu(self.dummy_fnu)
        self.assertTrue(np.isclose(flux, 3631.0 * u.Jy),
                        f"Unexpected integrated flux value: {flux}")

        mag, _ = filter.get_ab(self.dummy_fnu)
        self.assertTrue(np.isclose(mag, 0.0, atol=1e-4),
                        f"Unexpected magnitude value: {mag}")

        fig = filter.plot(show=False)

    def test_equivalent_width(self):
        eqwidth = observables.EquivalentWidth.from_name("lick_ha")
        ew, ew_err = eqwidth.compute_ew(self.dummy_wavelength, self.dummy_flam)
        self.assertTrue(np.isfinite(ew), "Unexpected EW value")


if __name__ == '__main__':
    unittest.main()


# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.


import os
import sys
sys.path.insert(0, os.path.abspath(
    os.path.join(os.path.dirname(__file__), '../../src'))
    )


# Avoid build failures on read the docs
try:
    import numpy
except ImportError:
    from mock import MagicMock

    MOCK_MODULES = [
        'numpy',
        'astropy',
        'astropy.convolution',
        'astropy.io',
        'astropy.wcs',
        'matplotlib',
        'matplotlib.colors',
        'matplotlib.pyplot',
        'synphot',
        'scipy',
        'scipy.ndimage',
        'scipy.ndimage.interpolation',
        'scipy.optimize',
        'scipy.signal',
        "past",
        "past.utils"
    ]
    sys.modules.update((mod_name, MagicMock()) for mod_name in MOCK_MODULES)



# -- Project information -----------------------------------------------------

project = 'Population Synthesis Toolkit'
copyright = '2024, Ascasibar Y., Corcho-Caballero P., Jiménez-López D.'
author = 'Ascasibar Y., Corcho-Caballero P., Jiménez-López D.'


import pst
version = '.'.join(pst.__version__.split(".")[0:2])
# The full version, including alpha/beta/rc tags.
release = pst.__version__

source_suffix = '.rst'

# The encoding of source files.
#source_encoding = 'utf-8-sig'

# The master toctree document.
master_doc = 'index'



# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.doctest',
    'sphinx.ext.intersphinx',
    'sphinx.ext.todo',
    'sphinx.ext.coverage',
    'sphinx.ext.mathjax',
    'sphinx.ext.viewcode',
    'sphinx.ext.napoleon',

]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'sphinx_rtd_theme'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']


# -- Extension configuration -------------------------------------------------

# -- Options for intersphinx extension ---------------------------------------

# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {'python': ('https://docs.python.org/3/', None),
                       'astropy': ('https://docs.astropy.org/en/stable/', None),
                       'numpy': ('https://numpy.org/doc/stable/', None)}

# -- Options for todo extension ----------------------------------------------

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False
autodoc_member_order = 'bysource'

# Napoleon settings
napoleon_google_docstring = False
napoleon_numpy_docstring = True
napoleon_include_init_with_doc = True
napoleon_include_private_with_doc = True
napoleon_include_special_with_doc = True
napoleon_use_admonition_for_examples = False
napoleon_use_admonition_for_notes = False
napoleon_use_admonition_for_references = False
napoleon_use_ivar = False
napoleon_use_param = False
napoleon_use_rtype = False
napoleon_preprocess_types = False
napoleon_type_aliases = None
napoleon_attr_annotations = True


import unittest
import numpy as np

from pst import SSP
from pst.observables import TopHatFilter

class TestSSP(unittest.TestCase):

    @classmethod
    def setUpClass(self):
        print("Setting SSP model")
        self.ssp_model = SSP.PopStar(IMF='cha')
    
    def test_get_ssp_logedges(self):
        self.assertIsNotNone(self.ssp_model.get_ssp_logedges())

    def test_cut_wavelength(self):
        self.ssp_model.cut_wavelength(5000, 9000)
    
    def test_interpolate_sed(self):
        new_wl = np.arange(5000, 8000, 5.5)
        self.ssp_model.interpolate_sed(new_wl)
        self.assertEqual(self.ssp_model.wavelength.size, new_wl.size)
        self.assertEqual(self.ssp_model.L_lambda.shape[-1], new_wl.size)

    def test_regrid(self):
        new_ages = np.array([0.5, 1.0, 5.0])
        new_metallicities = np.array([0.01, 0.02])
        copy_ssp = self.ssp_model.copy()
        copy_ssp.regrid(new_ages, new_metallicities)

    def test_get_mass_lum_ratio(self):
        mass_to_lum = self.ssp_model.get_mass_lum_ratio([5000, 5500])
        self.assertTrue(np.isfinite(mass_to_lum).all(),
                        "Mass-to-light ratio is NaN")

    def test_compute_photometry(self):
        filters = [TopHatFilter(central_wave=cw, width=w,
                                wavelength=self.ssp_model.wavelength
                                ) for cw, w in zip(
            [2000, 4000, 6000], [10, 100, 1000])]
        photometry = self.ssp_model.compute_photometry(filters, z_obs=0.0)
        self.assertTrue(np.isfinite(photometry).all(), "SSP photometry is NaN")

if __name__ == '__main__':
    unittest.main()


import os
from glob import glob
from copy import deepcopy
import numpy as np

from astropy.io import fits
from astropy.wcs import WCS
from astropy.table import Table
from astropy import units as u
from astropy import units

from pst.utils import check_unit, flux_conserving_interpolation

class SSPBase(object):
    """Base class that represents a model of Simple Stellar Populations.
    
    Description
    -----------
    This class is meant for representing a model of Simple Stellar Populations
    that is mainly composed by a discreted grid of ages and metallicities and
    their associated spectral energy distributions.

    Attributes
    ----------
    ages: astropy.units.Quantity
        Ages of the SSPs.
    metallicities: astropy.units.Quantity
        Metallicities of the SSPs.
    L_lambda: astropy.units.Quantity
        Spectral energy distribution of each SSP. Each dimension correspond to
        (metallicity, ages, wavelength).
    wavelength: astroy.units.Quantity
        Wavelength array associated to the SED of the SSPs.
    """
    default_path = os.path.join(os.path.dirname(__file__), "data", "ssp")

    @property
    def ages(self):
        return self._ages

    @ages.setter
    def ages(self, ages_array):
        if not isinstance(ages_array, units.Quantity):
            raise NameError("Ages must be an astropy.Quantity")
        else:
            self._ages = ages_array

    @property
    def metallicities(self):
        return self._metallicities

    @metallicities.setter
    def metallicities(self, metallicities_array):
        if not isinstance(metallicities_array, units.Quantity):
            raise NameError("metallicities must be an astropy.Quantity")
        else:
            self._metallicities = metallicities_array

    @property
    def L_lambda(self):
        return self._L_lambda

    @L_lambda.setter
    def L_lambda(self, l_lamdba):
        if not isinstance(l_lamdba, units.Quantity):
            raise NameError("L_lambda must be an astropy.Quantity")
        else:
            self._L_lambda = l_lamdba

    @property
    def wavelength(self):
        return self._wavelength
    
    @wavelength.setter
    def wavelength(self, wave):
        if not isinstance(wave, units.Quantity):
            raise NameError("wavelength must be an astropy.Quantity")
        else:
            self._wavelength = wave
       
    def get_weights(self, ages, metallicities, masses=None):
        """2D interpolation of a list of ages and metallicities.
        
        Parameters
        ----------
        ages : np.array or astropy.units.Quantity
            SSP ages to interpolate.
        metallicites : np.array or astropy.units.Quantity
            Metallicity associated to each age.
        masses : np.array, astropy.units.Quantity or None, optional
            Stellar mass corresponding to each SSP.
        """
        ages = check_unit(ages, u.Gyr)
        metallicities = check_unit(metallicities, u.dimensionless_unscaled)
        if masses is None:
            masses = np.ones(ages.size) << u.Msun
        else:
            masses = check_unit(masses, u.Msun)

        age_idx = np.clip(self.ages.searchsorted(ages), 1, self.ages.size-1)
        weights_age = np.log(ages / self.ages[age_idx-1])
        weights_age /= np.log(self.ages[age_idx] / self.ages[age_idx-1])
        weights_age = np.clip(weights_age, 0., 1.)

        z_idx = np.clip(self.metallicities.searchsorted(metallicities), 1, self.metallicities.size-1)
        weights_z = np.log(metallicities / self.metallicities[z_idx-1])
        weights_z /= np.log(self.metallicities[z_idx] / self.metallicities[z_idx-1])
        weights_z = np.clip(weights_z, 0., 1.)

        weights = np.zeros((self.metallicities.size, self.ages.size)) << masses.unit
        np.add.at(weights, (z_idx, age_idx), masses * weights_age * weights_z)
        np.add.at(weights, (z_idx-1, age_idx), masses * weights_age * (1-weights_z))
        np.add.at(weights, (z_idx-1, age_idx-1), masses * (1-weights_age) * (1-weights_z))
        np.add.at(weights, (z_idx, age_idx-1), masses * (1-weights_age) * weights_z)
        return weights

    def get_ssp_l_lambda(self, age, metallicity):
        """Compute the SED associated to an SSP of a given age and metallicity.
        
        Parameters
        ----------
        age : float or u.Quantity
            SSP age.
        metallicity : float or u.Quantity
            SSP metallicity
        
        Returns
        -------
        sed : u.Quantity
            Spectra energy distribution associated to the SSP.
        """
        age = check_unit(age, u.Gyr)
        metallicity = check_unit(metallicity, u.dimensionless_unscaled)

        age_idx = np.clip(self.ages.searchsorted(age), 1, self.ages.size-1)
        weights_age = np.log(age / self.ages[age_idx-1])
        weights_age /= np.log(self.ages[age_idx] / self.ages[age_idx-1])
        weights_age = np.clip(weights_age, 0., 1.)

        z_idx = np.clip(self.metallicities.searchsorted(metallicity), 1,
                        self.metallicities.size-1)
        weights_z = np.log(metallicity / self.metallicities[z_idx-1])
        weights_z /= np.log(self.metallicities[z_idx] / self.metallicities[z_idx-1])
        weights_z = np.clip(weights_z, 0., 1.)

        sed = self.L_lambda[z_idx, age_idx] *  weights_age * weights_z
        sed += self.L_lambda[z_idx-1, age_idx] * weights_age * (1-weights_z)
        sed += self.L_lambda[z_idx-1, age_idx-1] * (1-weights_age) * (1-weights_z)
        sed += self.L_lambda[z_idx-1, age_idx] * (1-weights_age) * weights_z
        return sed

    def get_ssp_logedges(self):
        """Get the edges of the SSP metallicities and ages."""
        logages = np.log10(self.ages / units.yr)
        lim = 1.5 * logages[[0, -1]] - 0.5 * logages[[1, -2]]
        logage_bin_edges = np.hstack(
                [lim[0], (logages[1:] + logages[:-1])/2, lim[1]])

        log_met_bins = np.log10(self.metallicities)
        lim = 1.5 * log_met_bins[[0, -1]] - 0.5 * log_met_bins[[1, -2]]
        logmet_bin_edges = np.hstack(
                [lim[0], (log_met_bins[1:] + log_met_bins[:-1])/2, lim[1]])
        return logmet_bin_edges, logage_bin_edges

    def regrid(self, age_bins, metallicity_bins, verbose=True):
        """Interpolate the SSP model to a new grid of input ages and metallicities.
        
        Parameters
        ----------
        age_bins : np.array or astropy.units.Quantity
        """
        if verbose:
            print("[SSP] Interpolating the SSP model to a new grid of ages and metallicities")
        age_bins = check_unit(age_bins, u.Gyr)
        metallicity_bins = check_unit(metallicity_bins, u.dimensionless_unscaled)
        # Bin the SED of the SSPs
        new_l_lambda = np.zeros((metallicity_bins.size, age_bins.size,
                                 self.wavelength.size)) * self.L_lambda.unit
        if verbose:
            print("New SSP SED shape: ", new_l_lambda.shape)
        for j, m_bin in enumerate(metallicity_bins):
            for i, a_bin in enumerate(age_bins):
                weights = self.get_weights(a_bin, m_bin, 1.0 * u.Msun)
                new_l_lambda[j, i] = np.sum(
                    self.L_lambda * weights[:, :, np.newaxis] / u.Msun, axis=(0, 1))

        if verbose:
            print("Updating SSP model metallicities, ages and SED")
        self.metallicities = metallicity_bins
        self.ages = age_bins
        self.L_lambda = new_l_lambda

    def cut_wavelength(self, wl_min=None, wl_max=None, verbose=True):
        """Cut model wavelength edges.
        
        Parameters
        ----------
        wl_min : float or astropy.units.Quantity, optional
            Minimum wavelength value.
        wl_max : float or astropy.units.Quantity, optional
            Maximum wavelength value.
        """
        if wl_min is None:
            wl_min = self.wavelength[0]
        else:
            wl_min = check_unit(wl_min, self.wavelength.unit)
        if wl_max is None:
            wl_max = self.wavelength[0]
        else:
            wl_max = check_unit(wl_max, self.wavelength.unit)

        cut_pts = np.where((self.wavelength >= wl_min) &
                           (self.wavelength <= wl_max))[0]
        if len(cut_pts) == 0:
            raise NameError(
                'Wavelength cuts {}, {} out of range for array with lims {} {}'
                .format(wl_min, wl_max, self.wavelength.min(),
                        self.wavelength.max())
                            )
        else:
            self.wavelength = self.wavelength[cut_pts]
            self.L_lambda = self.L_lambda[:, :, cut_pts]
            if verbose:
                print('[SSP] Models cut between {} {}'.format(wl_min, wl_max))

    def interpolate_sed(self, new_wl, verbose=True, log=False, **interp_kwargs):
        """Flux-conserving interpolation.

        Parameters
        ----------
        - new_wl: bin centers of the new interpolated points.
        """
        if not isinstance(new_wl, units.Quantity):
            new_wl = new_wl << self.wavelength.unit

        if verbose:
            print('[SSP] Interpolating SSP SEDs')

        if log:
            target_wl = np.log(new_wl.to_value(self.wavelength.unit))
            ref_wl = np.log(self.wavelength.value)
        else:
            target_wl = new_wl
            ref_wl = self.wavelength

        new_l_lambda = np.empty(
            shape=(self.metallicities.size, self.ages.size,
                   new_wl.size), dtype=np.float32) * self.L_lambda.unit

        for i in range(self.L_lambda.shape[0]):
            for j in range(self.L_lambda.shape[1]):
                new_l_lambda[i, j] = flux_conserving_interpolation(
                    target_wl, ref_wl, self.L_lambda[i, j],
                    **interp_kwargs)

        self.L_lambda = new_l_lambda
        self.wavelength = new_wl

    def get_mass_lum_ratio(self, wl_range):
        """Compute the mass-to-light ratio within a wavelength range."""
        if not isinstance(wl_range, u.Quantity):
            print("Assuming that input wavelength range has same units as wavelength")
            wl_range = np.array(wl_range) * self.wavelength.unit
        pts = np.where((self.wavelength >= wl_range[0]) &
                       (self.wavelength <= wl_range[1]))[0]
        mass_to_lum = np.empty(
            (self.metallicities.size, self.ages.size)
            ) * 1 / (self.L_lambda.unit * self.wavelength.unit)

        for i in range(self.metallicities.size):
            for j in range(self.ages.size):
                mass_to_lum[i, j] = 1/np.mean(self.L_lambda[i, j][pts] * self.wavelength[pts])
        return mass_to_lum
    
    def get_specific_mass_lum_ratio(self, wl_range):
        """Compute the mass-to-light ratio per wavelength unit within a wavelength range."""
        if not isinstance(wl_range, u.Quantity):
            print("Assuming that input wavelength range has same units as wavelength")
            wl_range = np.array(wl_range) * self.wavelength.unit
        pts = np.where((self.wavelength >= wl_range[0]) &
                       (self.wavelength <= wl_range[1]))[0]
        mass_to_lum = np.empty(
            (self.metallicities.size, self.ages.size)
            ) * 1 / self.L_lambda.unit

        for i in range(self.metallicities.size):
            for j in range(self.ages.size):
                mass_to_lum[i, j] = 1/np.mean(self.L_lambda[i, j][pts])
        return mass_to_lum
    
    def compute_photometry(self, filter_list, z_obs=0.0, verbose=True):
        """Compute the SSP synthetic photometry of a set of filters.
        
        Parameters
        ----------
        filter_list: list of pst.observable.Filter
            A list of photometric filters.

        Returns
        -------
        photometry: np.ndarray
            Array storing the photometry. The dimensions correspond to
            filter, metallicity and age.
        """
        if verbose:
            print("Computing synthetic photometry for SSP model")
        self.photometry = np.zeros((len(filter_list),
                                    *self.L_lambda.shape[:-1])) * u.Jy / u.Msun
        self.photometry_filters = filter_list
        for ith, f in enumerate(filter_list):
            f.interpolate(self.wavelength * (1 + z_obs))
            flux, _ = f.get_fnu(
                    self.L_lambda  * u.Msun / 4 / np.pi / (10 * u.pc)**2,
                    mask_nan=False)
            flux /= 1 + z_obs
            self.photometry[ith] = flux.to('Jy') / u.Msun
        return self.photometry

    def copy(self):
        """Return a copy of the SSP model."""
        return deepcopy(self)


class PopStar(SSPBase):
    """
    PopStar SSP models (Mollá et al. 2009).

    This class represents the PopStar evolutionary synthesis models for simple
    stellar populations as presented in Mollá et al. (2009). It provides
    access to spectral energy distributions (SEDs) across a range of metallicities
    and ages, optionally including nebular emission.

    Parameters
    ----------
    IMF : str
        Initial Mass Function name. Must be one of the supported IMFs listed below:

        - 'sal' : Salpeter (1955)
        - 'fer' : Ferrini, Penco & Palla (1990)
        - 'kro' : Kroupa (2002)
        - 'cha' : Chabrier (2003)

        This keyword is used to select the corresponding PopStar model files.
    nebular : bool, optional
        If True, include nebular emission in the SEDs. Default is False, meaning
        only the stellar continuum is considered.
    path : str or None, optional
        Filesystem path to the PopStar model data. If None (default), the package
        default path plus 'PopStar' subdirectory is used.
    verbose : bool, optional
        If True (default), print informational messages during initialization.

    Example
    -------
    Initialize with the desired IMF and nebular option:

    >>> from pst.SSP import PopStar
    >>> help(PopStar) # Retrieve the class documentation
    >>> popstar = PopStar(IMF="cha", nebular=True) # Initialise the model
    >>> print(popstar.wavelength)
    <Quantity [91. , 94. , 96. , ..., 1200000., 1400000., 1600000.] Angstrom>
    >>> oldest_sed = popstar.L_lambda[-1, -1] # Get the SED of the oldest SSP with highest metallicity

    Notes
    -----
    - The PopStar models cover metallicities from 0.0001 to 0.05 and ages spanning from
      10^5 to ~10^10 years.
    - Nebular emission is included only if `nebular=True` is set.
    - Model files are expected to be in compressed FITS format with specific naming
      conventions as per Mollá et al. 2009.

    For more details see the :class:`SSPBase` base class documentation.

    References
    ----------
    Mollá, M., García-Vargas, M. L., & Bressan, A. (2009). PopStar I: evolutionary synthesis model description. `MNRAS, 398(1), 451-470 <https://academic.oup.com/mnras/article/398/1/451/1099698>`_.
    `Fractal data repository <https://www.fractal-es.com/PopStar/>`_
    """
    def __init__(self, IMF, nebular=False, path=None, verbose=True):
        if path is None:
            self.path = os.path.join(self.default_path, 'PopStar')
        else:
            self.path = path
        self.metallicities = np.array([0.0001, 0.0004, 0.004, 0.008, 0.02,
                                       0.05]) * units.dimensionless_unscaled
        self.log_ages_yr = np.array([5.00, 5.48, 5.70, 5.85, 6.00, 6.10, 6.18,
                                     6.24, 6.30, 6.35, 6.40, 6.44, 6.48, 6.51,
                                     6.54, 6.57, 6.60, 6.63, 6.65, 6.68, 6.70,
                                     6.72, 6.74, 6.76, 6.78, 6.81, 6.85, 6.86,
                                     6.88, 6.89, 6.90, 6.92, 6.93, 6.94, 6.95,
                                     6.97, 6.98, 6.99, 7.00, 7.04, 7.08, 7.11,
                                     7.15, 7.18, 7.20, 7.23, 7.26, 7.28, 7.30,
                                     7.34, 7.38, 7.41, 7.45, 7.48, 7.51, 7.53,
                                     7.56, 7.58, 7.60, 7.62, 7.64, 7.66, 7.68,
                                     7.70, 7.74, 7.78, 7.81, 7.85, 7.87, 7.90,
                                     7.93, 7.95, 7.98, 8.00, 8.30, 8.48, 8.60,
                                     8.70, 8.78, 8.85, 8.90, 8.95, 9.00, 9.18,
                                     9.30, 9.40, 9.48, 9.54, 9.60, 9.65, 9.70,
                                     9.74, 9.78, 9.81, 9.85, 9.90, 9.95, 10.00,
                                     10.04, 10.08, 10.11, 10.12, 10.13, 10.14,
                                     10.15, 10.18]) * units.dimensionless_unscaled
        self.ages = 10**self.log_ages_yr * units.yr
        if verbose:
            print("> Initialising Popstar models (IMF='"+IMF+"')")
        # isochrone age in delta [log(tau)]=0.01
        if nebular:
            column = "_total"
            if verbose:
                print('--> Including NEBULAR emission')
        else:
            column = "_stellar"
            if verbose:
                print('--> Only stellar continuum')

        with fits.open(
            os.path.join(self.path, f"popstar_{IMF.lower()}_0.15_100.fits.gz")
            ) as hdul:
            self.wavelength = hdul[1].data["wavelength"] * u.Unit(
                hdul[1].header["TUNIT1"])

            self.L_lambda = np.empty(
                shape=(self.metallicities.size, self.log_ages_yr.size,
                    self.wavelength.size), dtype=np.float32
                    ) * u.Unit(hdul[2].header["TUNIT1"])

            for i, Z in enumerate(self.metallicities.value):
                table = hdul["SED_Z_0.{:04.0f}".format(Z*1e4)].data
                for j, age in enumerate(self.log_ages_yr.value):
                    self.L_lambda[i][j] = table[
                        "logage_yr_{:02.2f}".format(age)
                        + column] * self.L_lambda.unit


class PyPopStar(SSPBase):
    """PyPopStar SSP models (Millán-Irigoyen et al. 2021).

    This class represents the PyPopStar simple stellar population (SSP) models
    as described in Millán-Irigoyen et al. (2021). It provides spectral energy
    distributions (SEDs) for a range of metallicities and ages, optionally
    including nebular emission.

    Parameters
    ----------
    IMF : str
        Initial Mass Function identifier. Must be one of the supported IMFs:

        - 'KRO' : Kroupa (2002)
        - 'SAL' : Salpeter (1955)
        - 'CHA' : Chabrier (2003)
        This keyword selects the corresponding model subdirectory and file set.
    nebular : bool, optional
        If True, nebular emission is included in the SEDs. Default is False,
        meaning only the stellar continuum is included.
    path : str or None, optional
        Filesystem path to the PyPopStar model data. If None (default), the
        package default path plus 'PyPopStar' and IMF subdirectories are used.
    verbose : bool, optional
        If True (default), prints informational messages during initialization.

    Example
    -------
    Initialize the PyPopStar model with a chosen IMF and nebular emission option:

    >>> from pst.SSP import PyPopStar
    >>> psp = PyPopStar(IMF='KRO', nebular=False)
    >>> print(psp.wavelength)
    <Quantity [ ... ] Angstrom>
    >>> sed_example = psp.L_lambda[0, 0]  # SED for lowest metallicity, youngest age

    Notes
    -----
    - PyPopStar models cover metallicities from 0.004 to 0.05 and ages spanning
      approximately 10^5 to 10^10 years.
    - Nebular emission is included only if `nebular=True`.
    - Model files must follow the naming convention:
      `'SSP-{IMF}_Z{metallicity}_logt{log_age}.fits'`.

    For additional details, see the :class:`SSPBase` base class documentation.

    References
    ----------
    Millán-Irigoyen, I.; Mollá, M.; Cerviño, M.; et al. (2021). HR-PYPOPSTAR: high-wavelength-resolution stellar populations evolutionary synthesis model `MNRAS, 506 (4), 4781-4799 <https://ui.adsabs.harvard.edu/abs/2021MNRAS.506.4781M/abstract>`_.

    `Fractal data repository <https://www.fractal-es.com/PopStar/>`_
    """
    def __init__(self, IMF, nebular=False, path=None, verbose=True):
        if path is None:
            self.path = os.path.join(self.default_path, 'PyPopStar', IMF)
        else:
            self.path = path
        self.metallicities = np.array([0.004, 0.008, 0.02, 0.05]
                                      ) * units.dimensionless_unscaled
        self.log_ages_yr = np.array([
        5.,  5.48,  5.7 ,  5.85,  6.  ,  6.1 ,  6.18,  6.24,  6.3 ,
        6.35,  6.4 ,  6.44,  6.48,  6.51,  6.54,  6.57,  6.6 ,  6.63,
        6.65,  6.68,  6.7 ,  6.72,  6.74,  6.76,  6.78,  6.81,  6.85,
        6.86,  6.88,  6.89,  6.9 ,  6.92,  6.93,  6.94,  6.95,  6.97,
        6.98,  6.99,  7.  ,  7.04,  7.08,  7.11,  7.15,  7.18,  7.2 ,
        7.23,  7.26,  7.28,  7.3 ,  7.34,  7.38,  7.41,  7.45,  7.48,
        7.51,  7.53,  7.56,  7.58,  7.6 ,  7.62,  7.64,  7.66,  7.68,
        7.7 ,  7.74,  7.78,  7.81,  7.85,  7.87,  7.9 ,  7.93,  7.95,
        7.98,  8.  ,  8.3 ,  8.48,  8.6 ,  8.7 ,  8.78,  8.85,  8.9 ,
        8.95,  9.  ,  9.18,  9.3 ,  9.4 ,  9.48,  9.54,  9.6 ,  9.65,
        9.7 ,  9.74,  9.78,  9.81,  9.85,  9.9 ,  9.95, 10.  , 10.04,
       10.08, 10.11, 10.12, 10.13, 10.14, 10.15, 10.18]
       ) * units.dimensionless_unscaled
        self.ages = 10**self.log_ages_yr  * units.yr
        # isochrone age in delta [log(tau)]=0.01
        # self.wavelength = np.loadtxt(os.path.join(self.path, 'KRO', 'sp',
        #                                           'sp_z0.004_logt05.00.dat'),
        #                              dtype=float, usecols=(0,), unpack=True
        #                              ) * u.Angstrom
        header = 'SSP-{}'.format(IMF)
        if nebular:
            if verbose:
                print("> Initialising Popstar models (neb em) (IMF='"
                      + IMF + "')")
            column = 'flux_total'
        else:
            if verbose:
                print("> Initialising Popstar models (no neb em) (IMF='"
                      + IMF + "')")
            column = 'flux_stellar'
        with fits.open(os.path.join(self.path, header+'_Z{:03.3f}_logt{:05.2f}.fits'.format(
                self.metallicities.value[0], self.log_ages_yr.value[0]))
                       ) as hdul:
            self.wavelength = hdul[1].data['wavelength'] * units.Angstrom # Angstrom

        self.L_lambda = np.empty(
            shape=(self.metallicities.size, self.log_ages_yr.size,
                   self.wavelength.size), dtype=np.float32) * u.Lsun / u.Angstrom / u.Msun

        for i, Z in enumerate(self.metallicities.value):
            for j, age in enumerate(self.log_ages_yr.value):
                filename = header+'_Z{:03.3f}_logt{:05.2f}.fits'.format(Z, age)
                file = os.path.join(self.path, filename)
                with fits.open(file) as hdul:
                    self.L_lambda[i][j] = hdul[1].data[column] * self.L_lambda.unit
                    hdul.close()
        # Avoid 0 flux
        self.L_lambda[self.L_lambda <= 0] += self.L_lambda[self.L_lambda > 0].min()


class BC03_2003(SSPBase):
    """
    BC03 SSP models (Bruzual & Charlot 2003, original version).

    This class implements the 2003 version of the Bruzual & Charlot (BC03)
    evolutionary synthesis models. It supports both the STELIB and BaSeL stellar
    spectral libraries and various IMFs.

    Parameters
    ----------
    isochrone : str, optional
        Isochrone set to use:'Padova1994' (default) or 'Padova2000'.
    model : str, optional
        Stellar spectral library. Must be one of:

        - 'BaSeL'  : Low-resolution spectra
        - 'STELIB' : High-resolution spectra

        Case-insensitive. Default is 'BaSeL'.
    imf : str, optional
        Initial Mass Function. Must be one of:

        - 'chabrier' or 'cha' : Chabrier (2003)
        - 'salpeter' or 'sal' : Salpeter (1955)
        - 'kroupa' or 'kro'    : Kroupa (2002)

        Case-insensitive. Default is 'chabrier'.
    path : str or None, optional
        Filesystem path to the BC03 model data. If None (default), the package
        default path is used.
    verbose : bool, optional
        If True (default), print informational messages during initialization.

    Example
    -------
    Create a BC03 SSP object using the Chabrier IMF and BaSeL library:

    >>> from pst.SSP import BC03_2003
    >>> ssp = BC03_2003(model='BaSeL', imf='Chabrier')
    >>> print(ssp.wavelength)
    <Quantity [ ... ] Angstrom>
    >>> sed = ssp.L_lambda[0, -1]  # SED for lowest metallicity and oldest age

    Notes
    -----
    - This implementation loads the original 2003 release of BC03.
    - Supported metallicities (Padova1994 isochrones):

        - 'm22': Z = 0.0001
        - 'm32': Z = 0.0004
        - 'm42': Z = 0.004
        - 'm52': Z = 0.008
        - 'm62': Z = 0.02
        - 'm72': Z = 0.05

    - SEDs are read from FITS files with filenames of the form:
      `bc2003_{res}_{metallicity}_{imf}_ssp.fits`
    - Units for `L_lambda` are Lsun / Angstrom / Msun.
    - Ages are loaded from a standard time grid provided in `TIME_SCALE.DAT`.

    See also
    --------
    :class:`SSPBase` : Base class for all SSP models.

    References
    ----------
    Bruzual, G. & Charlot, S. (2003).
    Stellar population synthesis at the resolution of 2003.
    `MNRAS, 344(4), 1000–1028 <https://ui.adsabs.harvard.edu/abs/2003MNRAS.344.1000B/abstract>`_.

    `Gustavo Bruzual page <https://www.bruzual.org/>`_
    """
    _metallicity_map = {
     # Padova1994
    'm22': 0.0001,
    'm32': 0.0004,
    'm42': 0.004,
    'm52': 0.008,
    'm62': 0.02,
    'm72': 0.05,
    #'m82': 0.1  Not available in the 2003 version
    }
    _resolution = {'BaSeL': 'lr', 'stelib': 'hr'}
    
    def __init__(self, isochrone='Padova1994', model='BaSeL',
                 imf='Chabrier', path=None, verbose=True) -> None:
        self.isochrone = isochrone
        self.model, model_key = self._parse_model(model)
        self.imf, imf_key = self._parse_imf(imf)

        if path is None:
            self.path = os.path.join(self.default_path,
                                     'BC03', 'bc03_2003ver', "bc03",
                                     self.isochrone, self.imf)
        else:
            self.path = path

        if verbose:
            print(f"> Initialising BC03 model {self.model} (IMF={self.imf})")
        self.metallicities = np.array(list(self._metallicity_map.values())
                                      ) * units.dimensionless_unscaled
        self.ages = np.loadtxt(
            os.path.join(self.default_path, 'BC03', 'TIME_SCALE.DAT')
            ) * units.yr

        self.log_ages_yr = np.log10(self.ages / units.yr)

        load_wavelength = False
        for i, metallicity_key in enumerate(self._metallicity_map.keys()):
            fits_path = os.path.join(
                self.path, f"bc2003_{self._resolution[model_key]}_{metallicity_key}_{imf_key}_ssp.fits")
            table = Table.read(fits_path)
            if not load_wavelength:
                self.wavelength = table['wavelength'].value * u.angstrom
                self.L_lambda = np.zeros((self.metallicities.size, self.ages.size,
                                          self.wavelength.size))  * u.Lsun / u.Angstrom / u.Msun
                load_wavelength = True
            table.remove_column("wavelength")
            for j, column in enumerate(table.itercols()):
                self.L_lambda[i, j] = column.value * self.L_lambda.unit

    def _parse_model(self, model):
        if 'basel' in model.lower():
            model = 'BaSeL'
            key = 'BaSeL'
        elif 'stelib' in model.lower():
            model = 'stelib'
            key = 'stelib'
        else:
            raise NameError(f"Unrecognized model: {model}.\n"
                             + "Select Basel or Stelib")
        return model, key

    def _parse_imf(self, imf):
        if 'cha' in imf.lower():
            imf = 'chabrier'
            key = 'chab'
        elif 'sal' in imf.lower():
            imf = 'salpeter'
            key = 'salp'
        elif 'kro' in imf.lower():
            imf = 'kroupa'
            key = 'kro'
        else:
            raise NameError(f"Unrecognized IMF: {imf}.\n"
                             + "Select Chabrier, Salpeter or Kroupa")
        return imf, key


class BC03_2013(SSPBase):
    """
    BC03 SSP models (Bruzual & Charlot 2003, 2013 updated version).

    This class implements the updated 2013 version of the Bruzual & Charlot (BC03)
    simple stellar population (SSP) models, using improved coverage of high
    metallicities and updated isochrone input. It supports both the STELIB and
    BaSeL stellar spectral libraries and several initial mass functions (IMFs).

    Parameters
    ----------
    isochrone : str, optional
        Isochrone set to use. Currently only 'Padova1994' is supported (default).
    model : str, optional
        Stellar spectral library. Must be one of:

        - 'BaSeL'  : Low-resolution spectra
        - 'STELIB' : High-resolution spectra

        Case-insensitive. Default is 'BaSeL'.
    imf : str, optional
        Initial Mass Function. Must be one of:

        - 'chabrier' or 'cha' : Chabrier (2003)
        - 'salpeter' or 'sal' : Salpeter (1955)
        - 'kroupa' or 'kro'    : Kroupa (2002)

        Case-insensitive. Default is 'chabrier'.
    path : str or None, optional
        Filesystem path to the BC03 model data. If None (default), the package
        default path is used.
    verbose : bool, optional
        If True (default), print informational messages during initialization.

    Example
    -------
    Create a BC03 SSP object using the Chabrier IMF and BaSeL library:

    >>> from pst.SSP import BC03_2013
    >>> ssp = BC03_2013(model='BaSeL', imf='Chabrier')
    >>> print(ssp.wavelength)
    <Quantity [ ... ] Angstrom>
    >>> sed = ssp.L_lambda[0, -1]  # SED for lowest metallicity and oldest age

    Notes
    -----
    - This version extends the original 2003 release by including an additional
      high-metallicity track (Z = 0.1).
    - Supported metallicities (Padova1994 isochrones):

        - 'm22': Z = 0.0001
        - 'm32': Z = 0.0004
        - 'm42': Z = 0.004
        - 'm52': Z = 0.008
        - 'm62': Z = 0.02
        - 'm72': Z = 0.05
        - 'm82': Z = 0.1

    - File naming convention:
      `bc2003_{res}_{model}_{metallicity}_{imf}_ssp.fits`
    - Units for `L_lambda` are Lsun / Angstrom / Msun.
    - Ages are taken from the same `TIME_SCALE.DAT` file as BC03_2003.

    See also
    --------
    :class:`SSPBase` : Common functionality for SSP models.

    References
    ----------
    Bruzual, G. & Charlot, S. (2003).
    Stellar population synthesis at the resolution of 2003.
    `MNRAS, 344(4), 1000–1028 <https://ui.adsabs.harvard.edu/abs/2003MNRAS.344.1000B/abstract>`_.

    `Gustavo Bruzual page <https://www.bruzual.org/>`_
    """
    _metallicity_map = {
     # Padova1994
    'm22': 0.0001,
    'm32': 0.0004,
    'm42': 0.004,
    'm52': 0.008,
    'm62': 0.02,
    'm72': 0.05,
    'm82': 0.1}
    _resolution = {'BaSeL': 'lr', 'stelib': 'hr'}
    
    def __init__(self, isochrone='Padova1994', model='BaSeL',
                 imf='Chabrier', path=None, verbose=True) -> None:
        self.isochrone = isochrone
        self.model, model_key = self._parse_model(model)
        self.imf, imf_key = self._parse_imf(imf)

        if path is None:
            self.path = os.path.join(self.default_path,
                                     'BC03', 'bc03_2013ver', "bc03",
                                     self.isochrone, self.imf)
        else:
            self.path = path

        if verbose:
            print(f"> Initialising BC03 model {self.model} (IMF={self.imf})")
        self.metallicities = np.array(list(self._metallicity_map.values())
                                      ) * units.dimensionless_unscaled
        self.ages = np.loadtxt(
            os.path.join(self.default_path, 'BC03', 'TIME_SCALE.DAT')
            ) * units.yr

        self.log_ages_yr = np.log10(self.ages / units.yr)

        load_wavelength = False
        for i, metallicity_key in enumerate(self._metallicity_map.keys()):
            fits_path = os.path.join(
                self.path, f"bc2003_{self._resolution[model_key]}_{model_key}_{metallicity_key}_{imf_key}_ssp.fits")
            table = Table.read(fits_path)
            if not load_wavelength:
                self.wavelength = table['wavelength'].value * u.angstrom
                self.L_lambda = np.zeros((self.metallicities.size, self.ages.size,
                                          self.wavelength.size))  * u.Lsun / u.Angstrom / u.Msun
                load_wavelength = True
            table.remove_column("wavelength")
            for j, column in enumerate(table.itercols()):
                self.L_lambda[i, j] = column.value * self.L_lambda.unit

    def _parse_model(self, model):
        if 'basel' in model.lower():
            model = 'BaSeL'
            key = 'BaSeL'
        elif 'stelib' in model.lower():
            model = 'stelib'
            key = 'stelib'
        else:
            raise NameError(f"Unrecognized model: {model}.\n"
                             + "Select Basel or Stelib")
        return model, key

    def _parse_imf(self, imf):
        if 'cha' in imf.lower():
            imf = 'chabrier'
            key = 'chab'
        elif 'sal' in imf.lower():
            imf = 'salpeter'
            key = 'salp'
        elif 'kro' in imf.lower():
            imf = 'kroupa'
            key = 'kro'
        else:
            raise NameError(f"Unrecognized IMF: {imf}.\n"
                             + "Select Chabrier, Salpeter or Kroupa")
        return imf, key


class BC03_2016(SSPBase):
    """
    BC03 SSP models (Bruzual & Charlot 2003, 2016 updated version).

    This class implements the 2016 version of the Bruzual & Charlot (BC03)
    simple stellar population (SSP) models. It includes higher metallicity
    coverage and new spectral libraries, including MILES, STELIB, and BaSeL.
    This version uses updated file naming conventions and directory structure.

    Parameters
    ----------
    model : str, optional
        Stellar spectral library. Must be one of:

        - 'BaSeL'  : BaSeL 3.1 + ATLAS (low resolution)
        - 'STELIB' : STELIB + ATLAS (high resolution)
        - 'MILES'  : MILES + STELIB (high resolution)

        Case-insensitive. Default is 'BaSeL'.
    imf : str, optional
        Initial Mass Function. Must be one of:

        - 'chabrier' or 'cha' : Chabrier IMF
        - 'salpeter' or 'sal' : Salpeter IMF
        - 'kroupa' or 'kro'    : Kroupa IMF

        Case-insensitive. Default is 'Kroupa'.
    path : str or None, optional
        Filesystem path to the BC03 model data. If None (default), the package
        default path is used.
    verbose : bool, optional
        If True (default), print informational messages during initialization.

    Example
    -------
    Create a BC03 SSP object using the Chabrier IMF and BaSeL library:

    >>> from pst.SSP import BC03_2016
    >>> ssp = BC03_2016(model='BaSeL', imf='Chabrier')
    >>> print(ssp.wavelength)
    <Quantity [ ... ] Angstrom>
    >>> sed = ssp.L_lambda[0, -1]  # SED for lowest metallicity and oldest age

    Notes
    -----
    - This version supports extended metallicity coverage up to Z = 0.1.
    - The directory and file naming conventions follow:

        bc2003_{res}_{model_key}_{metallicity}_{imf_key}_ssp.fits

    - Supported metallicities (Padova1994 isochrones):

        - 'm22': Z = 0.0001
        - 'm32': Z = 0.0004
        - 'm42': Z = 0.004
        - 'm52': Z = 0.008
        - 'm62': Z = 0.02
        - 'm72': Z = 0.05
        - 'm82': Z = 0.1

    - MILES input is selected via `model='MILES'` (resolved as `xmiless` internally).

    See also
    --------
    :class:`SSPBase` : Base class for all SSP models.

    References
    ----------
    Bruzual, G., & Charlot, S. (2003).
    Stellar population synthesis at the resolution of 2003.
    `MNRAS, 344(4), 1000–1028 <https://ui.adsabs.harvard.edu/abs/2003MNRAS.344.1000B/abstract>`_.

    `Gustavo Bruzual page <https://www.bruzual.org/>`_

    Gutkin, J., Charlot, S., & Bruzual, G. (2016).
    Modelling the nebular emission from primeval to present-day star-forming galaxies
    `MNRAS, 462(2), 1757–1774 <https://ui.adsabs.harvard.edu/abs/2016MNRAS.462.1757G>`_.
    """

    _metallicity_map = {
     # Padova1994
    'm22': 0.0001,
    'm32': 0.0004,
    'm42': 0.004,
    'm52': 0.008,
    'm62': 0.02,
    'm72': 0.05,
    'm82': 0.1}
    _resolution = {'BaSeL': 'lr', 'stelib': 'hr', 'xmiless': 'hr'}
    
    def __init__(self, model='BaSeL', imf='Kroupa', path=None, verbose=True) -> None:
        self.model, model_key = self._parse_model(model)
        self.imf, imf_key = self._parse_imf(imf)

        if path is None:
            self.path = os.path.join(self.default_path, 'BC03', 'bc03_2016ver',
                                     self.model, self.imf)
        else:
            self.path = path

        if verbose:
            print(f"> Initialising BC03 model {self.model} (IMF={self.imf})")
        self.metallicities = np.array(list(self._metallicity_map.values())) * units.dimensionless_unscaled
        self.ages = np.loadtxt(
            os.path.join(self.default_path, 'BC03', 'TIME_SCALE.DAT')) * units.yr
        
        self.log_ages_yr = np.log10(self.ages / units.yr)

        load_wavelength = False
        for i, metallicity_key in enumerate(self._metallicity_map.keys()):
            fits_path = os.path.join(
                self.path, f"bc2003_{self._resolution[model_key]}_{model_key}_{metallicity_key}_{imf_key}_ssp.fits")
            table = Table.read(fits_path)
            if not load_wavelength:
                self.wavelength = table['wavelength'].value * u.angstrom
                self.L_lambda = np.zeros((self.metallicities.size, self.ages.size,
                                          self.wavelength.size))  * u.Lsun / u.Angstrom / u.Msun
                load_wavelength = True
            table.remove_column("wavelength")
            for j, column in enumerate(table.itercols()):
                self.L_lambda[i, j] = column.value * self.L_lambda.unit


    def _parse_model(self, model):
        if 'basel' in model.lower():
            model = 'BaSeL3.1_Atlas'
            key = 'BaSeL'
        elif 'stelib' in model.lower():
            model = 'Stelib_Atlas'
            key = 'stelib'
        elif 'miles' in model.lower():
            model = 'Miles_Atlas'
            key = 'xmiless'
        else:
            raise NameError(f"Unrecognized model: {model}.\n"
                             + "Select Basel, Stelib, Miles")
        return model, key

    def _parse_imf(self, imf):
        if 'cha' in imf.lower():
            imf = 'Chabrier_IMF'
            key = 'chab'
        elif 'sal' in imf.lower():
            imf = 'Salpeter_IMF'
            key = 'salp'
        elif 'kro' in imf.lower():
            imf = 'Kroupa_IMF'
            key = 'kroup'
        else:
            raise NameError(f"Unrecognized IMF: {imf}.\n"
                             + "Select Chabrier, Salpeter or Kroupa")
        return imf, key

class BaseGM(SSPBase):
    """
    SSP model class for the GM (Granada-MILES) base used in the CALIFA survey.

    This class loads a combined grid of simple stellar population (SSP) models:

    - The **young stellar populations** (ages < 63 Myr) are taken from GRANADA models developed by González Delgado et al. (2005).
    - The **older populations** (ages ≥ 63 Myr) are based on the MILES spectral
      library, specifically using the models from Vazdekis et al. (2010).

    Parameters
    ----------
    path : str or None, optional
        Path to the `gsd01_156.fits` file containing the GRANADA SSP model cube.
        If None, a default path is used. The companion `fits_like_properties.dat`
        file containing the age and metallicity grid must reside in the same directory.
    verbose : bool, optional
        If True, print initialization messages. Default is True.

    Notes
    -----
    - The Granada models are provided as 156 SEDs, reshaped to a (39 × 4) grid
      of ages and metallicities.
    - All models assume a **Salpeter (1955) initial mass function (IMF)**.
    - Spectral range: **3000 Å to 7000 Å**, with **0.3 Å resolution**.

    Example
    -------
    >>> from pst.SSP import BaseGM
    >>> gm = BaseGM(verbose=False)
    >>> print(gm.wavelength[0], gm.wavelength[-1])
    3000.0 Angstrom 7000.0 Angstrom
    >>> print(gm.ages.min(), gm.ages.max())
    1e6 yr 1.3e10 yr

    References
    ----------
    González Delgado, R. M., et al. (2005).
    "Evolutionary stellar population synthesis at high spectral resolution: optical wavelengths ." MNRAS, 357, 945.
    https://ui.adsabs.harvard.edu/abs/2005MNRAS.357..945G

    Sánchez, S. F., et al. (2012).
    "CALIFA, the Calar Alto Legacy Integral Field Area survey: I. Survey presentation."
    A&A, 538, A8. https://ui.adsabs.harvard.edu/abs/2012A%26A...538A...8S
    """
    def __init__(self, path=None, verbose=True):
        if path is None:
            self.path = os.path.join(self.default_path, 'BaseGM',
                                     'gsd01_156.fits')
            self._ssp_properties_path = os.path.join(
            self.default_path,
            'BaseGM', 'fits_like_properties.dat')
        else:
            self.path = path
            self._ssp_properties_path = os.path.join(os.path.dirname(path),
                                                    "fits_like_properties.dat")

        self.metallicities = np.loadtxt(self._ssp_properties_path, usecols=(1)
                                        ) * units.dimensionless_unscaled
        self.ages = np.loadtxt(self._ssp_properties_path, usecols=(0)) * units.yr
        self.log_ages_yr = np.log10(self.ages / units.yr)

        if verbose:
            print("> Initialising GRANADA models (IMF=Salpeter)")
        ssp_fits = fits.open(self.path)
        wl0 = ssp_fits[0].header['CRVAL1']
        deltawl = ssp_fits[0].header['CDELT1']

        self._header = ssp_fits[0].header
        self.norm = np.array(
            [ssp_fits[0].header['NORM'+str(i)] for i in range(156)]
            ) * units.Lsun / units.Msun / units.Angstrom

        SED = ssp_fits[0].data * self.norm[:, np.newaxis]
        SED = SED.reshape(39, 4, -1)
        self.norm = self.norm.reshape(39, 4)
        self.age_sort = np.argsort(self.ages.reshape(39, 4), axis=0)
        SED = SED[self.age_sort[:, 0], :, :]
        self.norm = self.norm[self.age_sort[:, 0], :]
        SED = SED.transpose(1, 0, 2)
        self.norm = self.norm.transpose(1, 0)
        self.wavelength = np.arange(wl0, wl0 + deltawl * SED.shape[-1],
                                    deltawl)  * units.Angstrom # AA
        self.L_lambda = SED
        ssp_fits.close()

        self.ages = np.unique(self.ages)
        self.log_ages_yr = np.log10(self.ages.value) * units.dimensionless_unscaled
        self.metallicities = np.sort(np.unique(self.metallicities))


# class FSPS(SSPBase):
#     """Fast Stellar Population Synthesis models..."""

#     def __init__(self):
#         print("> Initialising FSPS models")
#         self.path = os.path.join(config.path_to_ssp_models, 'FSPS',
#                                  'fsps_ssp_models.hdf5')
#         file = h5py.File(self.path, 'r')
#         elements = list(file.keys())
#         log_ages = []
#         metallicities = []
#         for elem_i in elements:
#             if elem_i.find('log_age') >= 0:
#                 init = 'log_age_'
#                 end = '_Z_'
#                 log_ages.append(elem_i[elem_i.find(init)+len(init):
#                                        elem_i.find(end)])
#                 metallicities.append(elem_i[elem_i.find(end)+len(end):])
#         metallicities = np.sort(np.unique(metallicities))
#         log_ages = np.unique(log_ages)
#         self.wavelength = file['wavelength'][()]
#         self.metallicities = np.array(metallicities, dtype=float)
#         self.log_ages_yr = np.sort(np.array(log_ages, dtype=float))
#         self.ages = 10**self.log_ages_yr

#         self.L_lambda = self.L_lambda = np.empty(
#             shape=(self.metallicities.size, self.log_ages_yr.size,
#                    self.wavelength.size), dtype=np.float32)
#         for i, met in enumerate(metallicities):
#             for j, log_age in enumerate(log_ages):
#                 name = 'log_age_{}_Z_{}'.format(log_age, met)
#                 l_lambda = file[name]['L_lambda'][()]
#                 self.L_lambda[i][j] = l_lambda
#         file.close()
#          = 'Lsun/Angstrom/Msun'


class XSL(SSPBase):
    """
    X-Shooter SSP empirical models.

    These models provide spectral energy distributions for simple stellar populations
    observed with the `X-Shooter instrument <http://www.eso.org/sci/facilities/paranal/instruments/xshooter/overview.html>`_. They cover a wide range of ages and
    metallicities, using two sets of isochrones and two IMFs.

    Parameters
    ----------
    IMF : str
        Initial Mass Function to use. Options:

        - 'Kroupa' : Kroupa (2001) IMF
        - 'Salpeter' : Salpeter (1955) IMF
    ISO : str
        Set of isochrones to use. Options:

        - 'P00' : Padova 2000 isochrones
        - 'PC' : PARSEC-COLIBRI isochrones
    path : str or None, optional
        Path to directory containing the SSP FITS files.
        If None, defaults to `default_path/XSL/IMF`.
    verbose : bool, optional
        If True, prints initialization messages. Default is True.

    Notes
    -----
    - Model files must be available in the specified `path`.

    References
    ----------
    Verro, K., Trager, S.C., Peletier, R.F., et al. 2022, `A&A, 661, A50 <https://ui.adsabs.harvard.edu/abs/2022A&A...661A..50V>`_.

    Project home page: http://xsl.u-strasbg.fr/index.html

    See also
    --------
    SSPBase
        Base class for SSP models.

    Example
    -------
    >>> from pst.SSP import XSL
    >>> xsl = XSL(IMF='Kroupa', ISO='P00', verbose=False)
    >>> print(xsl.wavelength[0], xsl.wavelength[-1])
    3000.0 Angstrom 25000.0 Angstrom
    >>> print(xsl.ages.min(), xsl.ages.max())
    5e7 yr 1.6e10 yr
    >>> print(xsl.metallicities)
    [0.0004 0.001 0.004 0.008 0.019 0.03]
    >>> print(xsl.L_lambda.shape)
    (6, 26, 58642)
    """
    _c_imf = dict(salpeter=9799552.50, kroupa=5567946.09)

    _initial_mass_functions = ["Kroupa", "Salpeter"]
    _iso_ages = {
            "PC": 10**np.arange(7.7, 10.1, 0.1) << u.yr,
            "P00": np.array([8.91e8, 1e9, 1.12e9, 1.26e9, 1.41e9, 1.58e9,
                             1.78e9, 2e9, 2.24e9, 2.51e9, 2.82e9, 3.16e9, 3.55e9,
                             3.98e9, 4.47e9, 5.01e9, 5.62e9, 6.31e9,
                             7.08e9, 7.94e9, 8.91e9, 1e10, 1.12e10, 1.26e10,
                             1.41e10, 1.59e10, 1.78e10]) << u.yr
            }
    _iso_metals = {
            "P00": np.array([0.0004, 0.001, 0.004, 0.008, 0.019, 0.03]
                           ) << u.dimensionless_unscaled,
            "PC": 10**np.array([-2.2, -2.0, -1.8, -1.6, -1.4, -1.2, -1.0, -0.8,
                             -0.6, -0.4, -0.2, 0, 0.2]) * 0.019 << u.dimensionless_unscaled
            }
    _n_wavelength = 58642

    def __init__(self, IMF, ISO, path=None, verbose=True):
        if verbose:
            print("> Initialising X-Shooter (XSL) models (IMF={}, ISO={})".format(
                IMF, ISO))
        if (IMF != 'Kroupa') & (IMF != 'Salpeter'):
            raise NameError('IMF not valid (use Kroupa | Salpeter)')
        if (ISO != 'P00') & (ISO != 'PC'):
            raise NameError('ISO not valid (use P00 for Padova2000 or PC for PARSEC/COLIBRI)')
        if path:
            self.path = path 
        else:
            self.path = os.path.join(self.default_path, 'XSL', IMF)

        self._get_ssps(ISO)
        self._fetch_files(ISO, IMF)
 
    def _get_ssps(self, iso):
        self.ages = self._iso_ages[iso]
        self.log_ages_yr = np.log10(self.ages.to_value("yr"))
        self.metallicities = self._iso_metals[iso]
        self.L_lambda = np.empty(
            shape=(self.metallicities.size, self.ages.size,
                   self._n_wavelength), dtype=np.float32
                   ) * units.Lsun / units.Msun / units.Angstrom
 
    def _fetch_files(self, iso, imf):
        c_solar = self._c_imf[imf.lower()]  # Convert to solar units
        if iso == "P00":
            file_fmt = r"XSL_SSP_T{:2.2e}_Z{}_Kroupa_P00.fits"
            for age_idx, age in enumerate(self.ages):
                for met_idx, met in enumerate(self.metallicities):
                    filename = file_fmt.format(age.to_value("yr"), met.value)
                    path_to_file = os.path.join(self.path, filename)
                    print(path_to_file)
                    if not os.path.isfile(path_to_file):
                        raise FileNotFoundError(f"{path_to_file} not found")
                    with fits.open(path_to_file) as hdul:
                        spec = hdul[0].data * c_solar
                        self.L_lambda[met_idx][age_idx] = spec * self._L_lambda.unit
        elif iso == "PC":
            file_fmt = r"XSL_SSP_logT{:.1f}_MH{:.1f}_Kroupa_PC.fits"
            for age_idx, age in enumerate(self.ages):
                for met_idx, met in enumerate(self.metallicities):
                    filename = file_fmt.format(np.log10(age.to_value("yr")),
                                               np.log10(met.value / 0.019))
                    if met == 0.019:
                        filename = file_fmt.format(np.log10(age.to_value("yr")),
                                               np.log10(met.value / 0.019 - 1e-3))
                    path_to_file = os.path.join(self.path, filename)
                    print(path_to_file)
                    if not os.path.isfile(path_to_file):
                        raise FileNotFoundError(f"{path_to_file} not found")
                    with fits.open(path_to_file) as hdul:
                        spec = hdul[0].data * c_solar
                        self.L_lambda[met_idx][age_idx] = spec * self._L_lambda.unit

        # Use the last file to load the wavelenth array
        with fits.open(path_to_file) as hdul:
            self.wavelength = 10**(
                (np.arange(0, hdul[0].data.size, 1) - hdul[0].header['CRPIX1'])
                * hdul[0].header['CDELT1'] + hdul[0].header['CRVAL1'] + 1
                ) * units.Angstrom

class EMILES(SSPBase):
    """
    E-MILES simple stellar population (SSP) models.

    These models provide high-quality empirical SSP spectra covering the
    spectral range 1680−50000 Å at moderately high resolution. The E-MILES
    library combines multiple stellar libraries and theoretical isochrones
    to deliver spectra of single-age, single-metallicity populations.

    The UV spectral range (1680–3540 Å) is computed using the `NGSL space-based
    stellar library <https://archive.stsci.edu/prepds/stisngsl/>`_, offering a significant improvement over earlier space-based
    models. The optical spectra use the MILES empirical library, and redder
    wavelengths are covered by `Indo-US <https://noirlab.edu/science/observing-noirlab/observing-kitt-peak/telescope-and-instrument-documentation/cflib>`_, `CaT <https://research.iac.es/proyecto/miles/pages/stellar-libraries/cat-library.php>`_, and `IRTF <https://irtfweb.ifa.hawaii.edu/~spex/IRTF_Spectral_Library/>`_ empirical libraries,
    all computed with consistent methods.

    The SSP spectra span metallicities from −1.79 < [M/H] < +0.26 and ages
    greater than 30 Myr, across several IMF types with varying slopes.

    Spectral resolution:

        - UV range (λ < 3060 Å): FWHM ≈ 3 Å
        - UV range (3060 Å < λ < 3540 Å): FWHM ≈ 5 Å
        - Optical range (3540 Å to 8950 Å): FWHM ≈ 2.5 Å
        - Infrared (longer wavelengths): σ = 60 km/s

    Parameters
    ----------
    iso : str
        Isochrone set to use. Options:

        - 'BASTI' : BASTI isochrones (Pietrinferni et al. 2004)
        - 'PADOVA00' : Padova 2000 isochrones (Girardi et al. 2000)
    imf : str
        Initial mass function to use. Options:

        - 'KROUPA_UNIVERSAL' : Kroupa universal IMF (Kroupa 2001)
        - 'UNIMODAL' : Unimodal IMF (single power-law, Vazdekis et al. 1996)
        - 'BIMODAL' : Bimodal IMF (Vazdekis et al. 1996)
        - 'CHABRIER' : Chabrier IMF (Chabrier 2003)
    path : str or None, optional
        Path to the directory containing the E-MILES FITS files.
        Defaults to `default_path/EMILES`.
    verbose : bool, optional
        If True, prints initialization messages. Default is True.

    Notes
    -----
    - The spectral resolution and wavelength coverage follow Vazdekis et al. (2016)
      and Röck et al. (2016).
    - The models cover a wavelength range of approximately 1680–50000 Å with
      spectral resolution ~2.5 Å (FWHM).
    - The metallicities are given as logarithmic values relative to solar,
      with solar metallicity fixed at Z=0.019.

    References
    ----------
    Vazdekis, A., Koleva, M., Ricciardelli, E., et al. 2016, `MNRAS, 463, 3409 <https://ui.adsabs.harvard.edu/abs/2016MNRAS.463.3409V/abstract>`_
    Röck, B., Vazdekis, A., Ricciardelli, E., et al. 2016, `A&A, 589, A73, 8 <https://ui.adsabs.harvard.edu/abs/2016A%26A...589A..73R/abstract>`_

    See also
    --------
    SSPBase
        Base class for simple stellar population models.

    Example
    -------
    >>> from pst.SSP import EMILES
    >>> emiles = EMILES(iso='BASTI', imf='KROUPA_UNIVERSAL', verbose=False)
    >>> print(emiles.wavelength[0], emiles.wavelength[-1])
    1680.0 Angstrom 50000.0 Angstrom
    >>> print(emiles.ages.min(), emiles.ages.max())
    30.0 Myr 14.0 Gyr
    >>> print(emiles.metallicities)
    [0.0001 0.00016 ... 0.03]  # actual values depend on the isochrone
    >>> print(emiles.L_lambda.shape)
    (12, 52, 53689)
    """
    _lib_isochrones = {"BASTI": "iTp",
                  "PADOVA00": "iPp"}
    _lib_imfs = {"KROUPA_UNIVERSAL": "ku1.30",
                "UNIMODAL": "un",
                "BIMODAL": "bi",
                "CHABRIER": "ch1.30"}

    
    _n_wavelength = 53689  # Spectra size

    _iso_logmetals = {
        "BASTI": np.array([-2.27, -1.79, -1.49, -1.26, -0.96, -0.66, -0.35,
                           -0.25, 0.06, 0.15, 0.26, 0.4]),
        "PADOVA00": np.array([-2.32, -1.71 , -1.31 , -0.71, -0.40, 0.00, 0.22])
    }
    _iso_ages = {
        "BASTI": np.array(
            [0.03, 0.04, 0.05, 0.06, 0.07, 0.08, 0.09, 0.10, 0.15, 0.20, 0.25,
            0.30, 0.35, 0.40, 0.45, 0.50, 0.60, 0.70, 0.80, 0.90, 1.0, 1.250,
            1.500, 1.750, 2.000, 2.25, 2.50, 2.75, 3.0, 3.25, 3.50, 3.75, 4.0,
            4.5, 5.0, 5.5, 6.0, 6.5, 7., 7.5, 8., 8.5, 9.0, 9.5, 10., 10.5, 11.,
            11.5, 12., 12.5, 13., 13.5, 14.]) << u.Gyr,
        "PADOVA00": np.array(
        [0.0631, 0.0708, 0.0794, 0.0891, 0.1000, 0.1122, 0.1259, 0.1413, 0.1585,
            0.1778, 0.1995, 0.2239, 0.2512, 0.2818, 0.3162, 0.3548, 0.3981, 0.4467,
            0.5012, 0.5623, 0.6310, 0.7079, 0.7943, 0.8913, 1.0000, 1.1220, 1.2589,
            1.4125, 1.5849, 1.7783, 1.9953, 2.2387, 2.5119, 2.8184, 3.1623, 3.5481,
            3.9811, 4.4668, 5.0119, 5.6234, 6.3096, 7.0795, 7.9433, 8.9125, 10.0000,
            11.2202, 12.5893, 14.1254, 15.8489, 17.7828]) << u.Gyr    
    }

    def __init__(self, iso, imf, path=None, verbose=True):
        if verbose:
            print("> Initialising E-MILES models (IMF={}, ISO={})".format(
                imf, iso))
        if path:
            self.path = path
        else:
            self.path = os.path.join(self.default_path, 'EMILES')
        model_name = self._get_models_prefix(iso, imf)
        self._load_models(model_name)

    def _get_models_prefix(self, iso, imf):
        """Get the file prefix used to load the model files."""
        model_name = os.path.join(self.path, "E")
        # IMF
        if not imf.upper() in self._lib_imfs.keys():
            raise NameError(f"Input IMF {imf} not recognized"
                            + f"\nThe available isochrones are: {self._lib_imfs.keys()}")
        else:
            model_name += self._lib_imfs[imf.upper()]
        # Metallicity and age
        model_name += r"Z{}{:.2f}T{:07.4f}_"

        # Isochrone
        if not iso.upper() in self._lib_isochrones:
            raise NameError(f"Input isochrone {iso} not recognized"
                            + f"\nThe available isochrones are: {list(self._lib_isochrones.keys())}")
        else:
            model_name += self._lib_isochrones[iso.upper()]
            self._ages = self._iso_ages[iso.upper()]
            self.log_ages_yr = np.log10(self._ages / units.yr)
            # Solar metallicity defined in V+16.
            self._logmetals = self._iso_logmetals[iso.upper()]
            self._metallicities = 10**self._logmetals * 0.019
        # Alpha over iron
        model_name += "0.00_baseFe.fits"
        return model_name

    def _load_models(self, model_name):
        """Load the SSP models from individual FITS files."""
        self.L_lambda = np.zeros(
            (self.metallicities.size, self.ages.size,
             self._n_wavelength))  * u.Lsun / u.Angstrom / u.Msun

        for met_idx, metal in enumerate(self._logmetals):
            if metal < 0:
                m_prefix = "m"
            else:
                m_prefix = "p"

            for age_idx, age in enumerate(self.ages.to_value("Gyr")):
                file = model_name.format(m_prefix, np.abs(metal), age)
                assert os.path.isfile(file), f"File {file} not found"
                with fits.open(file) as hdul:
                    self.L_lambda[met_idx, age_idx] = hdul[0].data << self.L_lambda.unit
        
        with fits.open(file) as hdul:
            wcs = WCS(hdul[0].header)
            self.wavelength = wcs.array_index_to_world_values(
                np.arange(0, self._n_wavelength)) << u.AA

if __name__ == '__main__':
    # ssp = PopStar(IMF='cha_0.15_100')
    from matplotlib import pyplot as plt
    ssp = BaseGM()

# %%                                                    ... Paranoy@ Rulz! ;^D
# Mr Krtxo \(ﾟ▽ﾟ)/
